from pathlib import Path
from typing import BinaryIO, List, Optional

import fastapi
import pluggy

import quetz
from quetz.condainfo import CondaInfo

hookspec = pluggy.HookspecMarker("quetz")


@hookspec
def register_router() -> 'fastapi.APIRouter':
    """add extra endpoints to the url tree.

    It should return an :py:class:`fastapi.APIRouter` with new endpoints definitions.
    By default it will be added to the root of the urlscheme"""


@hookspec
def post_add_package_version(
    version: 'quetz.db_models.PackageVersion', condainfo: 'quetz.condainfo.CondaInfo'
) -> None:
    """hook for post-processsing after adding a package file.

    :param quetz.db_models.PackageVersion version:
        package version model that was added in to the database

    :param quetz.condainfo.CondaInfo condainfo:
        metadata extracted from the archive

    """
    pass


@hookspec
def post_package_indexing(
    tempdir: Path,
    channel_name: str,
    subdirs: List[str],
    files: dict,
    packages: dict,
) -> None:
    """hook for post-processsing after building indexes.

    :param quetz.pkgstores.PackageStore pkgstore:
        package store used to store/retrieve packages

    :param str channel_name:
        metadata extracted from the archive

    :param list subdirs:
        list of subdirs with indexes

    :param dict files:
        a dict that contains list of files for each subdir
        - used in updating the index

    :param dict packages:
        a dict that contains list of packages for each subdir
        - used in updating the index
    """
    pass


@hookspec
def post_index_creation(
    raw_repodata: dict,
    channel_name: str,
    subdir: str,
) -> None:
    """hook for post-processsing after creating package index.

    :param dict raw_repodata:
        the package index

    :param str channel_name:
        the channel name

    :param str subdir:
        the subdirectory to which belongs the package index
    """
    pass


@hookspec
def validate_new_package(
    channel_name: str,
    package_name: str,
    file_handler: Optional[BinaryIO],
    condainfo: Optional[CondaInfo],
) -> None:
    """Validate new package name.

    It should raise :class:``quetz.errors.ValidationError`` if
    a package is not valid.

    :param str package_name:
        name of the package

    :param str channel_name:
        name of channel with the package

    :param BinaryIO file_handler:
       handler to the package file, it can be None if a package is created but
       no file was yet uploaded

    :param CondaInfo condainfo:
       CondaInfo instance with package metadata, it can be None if file was not
       uploaded
    """


@hookspec
def check_additional_permissions(db, user_id, user_role) -> bool:
    """
    Check if the user has appropriate permissions

    :param str user_id:
        id of the user
    """
    return True



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

from datetime import datetime
from enum import Enum

import sqlalchemy as sa

from quetz.db_models import UUID, Base


class JobStatus(str, Enum):
    pending = "pending"
    queued = "queued"
    running = "running"
    success = "success"
    failed = "failed"
    timeout = "timeout"
    cancelled = "cancelled"


class TaskStatus(str, Enum):
    created = "created"
    pending = "pending"
    running = "running"
    success = "success"
    failed = "failed"
    skipped = "skipped"


class ItemsSelection(Enum):
    watch = "watch"
    watch_for = "watch_for"
    all = "all"
    list = "list"


class Job(Base):
    __tablename__ = "jobs"

    id = sa.Column(sa.Integer, primary_key=True)
    created = sa.Column(sa.DateTime, nullable=False, default=datetime.utcnow)
    updated = sa.Column(sa.DateTime, nullable=False, default=datetime.utcnow)
    manifest = sa.Column(sa.LargeBinary(), nullable=False)
    owner_id = sa.Column(
        UUID, sa.ForeignKey("users.id", ondelete="cascade"), nullable=True
    )
    owner = sa.orm.relationship(
        "User", backref=sa.orm.backref("jobs", cascade="all,delete")
    )
    items_spec = sa.Column(sa.String(), nullable=True)
    items = sa.Column(
        sa.Enum(ItemsSelection),
        nullable=False,
        default=ItemsSelection.all,
    )
    status = sa.Column(
        sa.Enum(JobStatus),
        nullable=False,
        default=JobStatus.pending,
    )

    tasks = sa.orm.relationship(
        'Task', back_populates='job', cascade="all,delete-orphan", uselist=True
    )

    extra_args = sa.Column(sa.String(), nullable=True)

    start_at = sa.Column(sa.DateTime, nullable=True)
    repeat_every_seconds = sa.Column(sa.Integer, nullable=True)


class Task(Base):
    __tablename__ = "tasks"

    id = sa.Column(sa.Integer, primary_key=True)
    created = sa.Column(sa.DateTime, nullable=False, default=datetime.utcnow)
    updated = sa.Column(sa.DateTime, nullable=False, default=datetime.utcnow)
    # name = sa.Column(sa.Unicode(256), nullable=False)
    status = sa.Column(
        sa.Enum(TaskStatus),
        nullable=False,
        default=TaskStatus.created,
    )
    job_id = sa.Column(
        sa.Integer, sa.ForeignKey("jobs.id", ondelete="cascade"), nullable=False
    )
    job = sa.orm.relationship("Job", back_populates='tasks')
    package_version_id = sa.Column(
        UUID, sa.ForeignKey("package_versions.id", ondelete="cascade"), nullable=True
    )
    package_version = sa.orm.relationship(
        "PackageVersion",
        backref=sa.orm.backref("tasks", cascade="all,delete-orphan"),
    )

    def __repr__(self):
        if self.package_version:
            filename = self.package_version.filename
        else:
            filename = None
        return (
            f"Task(id={self.id}, package_version='{filename},"
            f" job_id={self.job_id}')"
        )



# Copyright 2020 QuantStack, Codethink Ltd
# Distributed under the terms of the Modified BSD License.

import json
from typing import Any, List, overload
from urllib.parse import quote

from quetz.config import Config, ConfigEntry, ConfigSection

from .oauth2 import OAuthAuthenticator


class JupyterConfigEntry:
    config_section = "jupyterhubauthenticator"
    registered_entries: List[ConfigEntry] = []
    config = None

    def __init__(self, dtype, default=None, required=True):
        self.dtype = dtype
        self.default = default
        self.required = required

    # these type annotations dont work yet, but I leave them for now
    # maybe someone will find a solution later
    # https://github.com/python/mypy/issues/2566#issuecomment-703998877
    @overload
    def __get__(self, instance: None, owner: Any) -> "JupyterConfigEntry":
        ...

    @overload
    def __get__(self, instance: object, owner: Any) -> str:
        ...

    def __get__(self, obj, objtype) -> str:
        return getattr(self.config, self.config_attr_name)

    def __set_name__(self, owner, name):
        self.attr_name = name
        self.config_attr_name = f"{self.config_section}_{name}"
        entry = ConfigEntry(
            name, self.dtype, default=self.default, required=self.required
        )
        self.registered_entries.append(entry)

    @classmethod
    def _make_config(cls):
        section = ConfigSection(
            cls.config_section,
            cls.registered_entries,
            required=False,
        )
        return [section]

    @classmethod
    def register(cls, config: Config):
        cls.config = config
        config_options = cls._make_config()
        config.register(config_options)
        return config.configured_section(cls.config_section)


class JupyterhubAuthenticator(OAuthAuthenticator):
    """Use Oauth2 protcol to authenticate with jupyterhub server, which acts
    as identity provider.

    To activate add the following section to the ``config.toml`` (see :ref:`configfile`):

    .. code::

      [jupyterhubauthenticator]

      # client credentials, they need to be registered with
      # jupyterhub by adding an external service
      client_id = "quetz_client"
      client_secret = "super-secret"

      # token enpoint of Jupyterhub, needs to be accessible from Quetz server
      access_token_url = "http://JUPYTERHUB_HOST:PORT/hub/api/oauth2/token"

      # authorize endpoint of JupyterHub, needs to be accessible from users' browser
      authorize_url = "http://JUPYTERHUB_HOST:PORT/hub/api/oauth2/authorize"

      # API root, needs to be accesible from Quetz server
      api_base_url = "http://JUPYTERHUB_HOST:PORT/hub/api/"

    To configure quetz as an oauth client in JupyterHub, you will need to define
    a `JupyterHub service <https://jupyterhub.readthedocs.io/en/stable/reference/services.html#externally-managed-services>`_. You can achieve it by adding the following to the
    ``jupyterhub_config.py`` file of your JupyterHub instance:

    .. code::

      c.JupyterHub.services = [
          {
              # service name, it will be used to setup routers
              'name': 'quetz',
              # quetz URL to setup redirections, only required if you use
              # JupyterHub url scheme
              'url': 'http://QUETZ_HOST:PORT',
              # any secret >8 characters, you will also need to set
              # the client_secret in the authenticator config with this
              # string
              'api_token': 'super-secret',
              # client_id in the authenticator config
              'oauth_client_id': 'quetz_client',
              # URL of the callback endpoint on the quetz server
              'oauth_redirect_uri': 'http://QUETZ_HOST:PORT/auth/jupyterhub/authorize',
          }
      ]
    """  # noqa

    provider = 'jupyterhub'

    # TODO: need to figure out how to use type annotations with descriptors
    # see also: https://github.com/python/mypy/pull/2266

    client_id = JupyterConfigEntry(str, required=True)  # type: ignore
    client_secret = JupyterConfigEntry(str, required=True)  # type: ignore

    access_token_url = JupyterConfigEntry(str, required=True)  # type: ignore
    validate_token_url = "authorizations/token/{}"
    authorize_url = JupyterConfigEntry(str, required=True)  # type: ignore
    api_base_url = JupyterConfigEntry(str, required=True)  # type: ignore

    client_kwargs = {
        "token_endpoint_auth_method": "client_secret_post",
        "token_placement": "uri",
    }

    async def userinfo(self, request, token):
        response = await self._get_user_for_token(token)
        profile = response.json()

        github_profile = {
            "id": profile["name"] + '_id',
            "name": profile["name"],
            "avatar_url": "",
            "login": profile["name"],
        }
        return github_profile

    async def _get_user_for_token(self, token):
        headers = {'Authorization': f'token {self.client_secret}'}
        access_token = quote(token['access_token'], safe='')

        # authlib client will be place token in query params
        # which are ignored by jupyterhub
        # this workaround is required to implement jupyterhub API
        # which puts the token as path parameter
        # https://jupyterhub.readthedocs.io/en/stable/_static/rest-api/index.html#path--authorizations-token--token-  # noqa
        resp = await self.client.get(
            f'authorizations/token/{access_token}', token=token, headers=headers
        )
        return resp

    async def validate_token(self, token):
        # access_token = json.loads(token)["access_token"]
        token = json.loads(token)
        resp = await self._get_user_for_token(token)
        return resp.status_code == 200

    def configure(self, config: Config):
        self.is_enabled = JupyterConfigEntry.register(config)

        super().configure(config)



"""py.test fixtures

Fixtures for Quetz components
-----------------------------
- `db`

"""
# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.


import uuid
from pathlib import Path

from pytest import fixture

from quetz.config import Config
from quetz.dao import Dao
from quetz.db_models import PackageVersion, Profile, User
from quetz.rest_models import Channel, Package

pytest_plugins = "quetz.testing.fixtures"


@fixture
def user_role():
    return None


@fixture
def user_without_profile(db, user_role):
    new_user = User(id=uuid.uuid4().bytes, username="bartosz", role=user_role)
    db.add(new_user)
    db.commit()

    yield new_user

    db.delete(new_user)
    db.commit()


@fixture
def user(db, user_without_profile):
    profile = Profile(
        name="Bartosz", avatar_url="http:///avatar", user=user_without_profile
    )
    db.add(profile)
    db.commit()

    yield user_without_profile

    db.query(Profile).filter(
        Profile.name == profile.name,
        Profile.avatar_url == profile.avatar_url,
        Profile.user_id == user_without_profile.id,
    ).delete()

    db.commit()


@fixture
def make_package_version(
    db,
    user,
    public_channel,
    channel_name,
    package_name,
    public_package,
    dao: Dao,
    config: Config,
):
    pkgstore = config.get_package_store()

    versions = []

    def _make_package_version(filename, version_number, platform="linux-64"):
        filename = Path(filename)
        with open(filename, "rb") as fid:
            pkgstore.add_file(fid.read(), channel_name, platform / filename)
        package_format = "tarbz2"
        package_info = "{}"
        version = dao.create_version(
            channel_name,
            package_name,
            package_format,
            platform,
            version_number,
            0,
            "",
            str(filename),
            package_info,
            user.id,
            size=11,
        )

        dao.update_package_channeldata(
            channel_name,
            package_name,
            {'name': package_name, 'subdirs': [platform]},
        )

        dao.update_channel_size(channel_name)

        versions.append(version)

        return version

    yield _make_package_version

    for version in versions:
        db.query(PackageVersion).filter(PackageVersion.id == version.id).delete()
    db.commit()


@fixture
def package_version(db, make_package_version):
    version = make_package_version("test-package-0.1-0.tar.bz2", "0.1")

    return version


@fixture
def public_channel(dao: Dao, user, channel_role, channel_name):
    channel_data = Channel(name=channel_name, private=False)
    channel = dao.create_channel(channel_data, user.id, channel_role)

    return channel


@fixture
def public_package(db, user, public_channel, dao, package_role, package_name):
    package_data = Package(name=package_name)

    package = dao.create_package(
        public_channel.name, package_data, user.id, package_role
    )

    return package


@fixture
def channel_name():
    return "my-channel"


@fixture
def package_name():
    return "my-package"


@fixture
def channel_role():
    return "owner"


@fixture
def package_role():
    return "owner"



# Copyright 2020 Codethink Ltd
# Distributed under the terms of the Modified BSD License.

import fnmatch
import hashlib
import json
import tarfile
import time
from io import BytesIO
from zipfile import ZipFile

import zstandard

from quetz import db_models

from .exceptions import PackageError

INFO_BOOLEAN_FIELDS = (
    "activate.d",
    "deactivate.d",
    "post_link",
    "pre_link",
    "pre_unlink",
    "binary_prefix",
    "text_prefix",
)
INFO_OPTIONAL_FIELDS = ("version",)
ABOUT_OPTIONAL_FIELDS = (
    "description",
    "dev_url",
    "doc_source_url",
    "doc_url",
    "home",
    "icon_hash",
    "icon_url",
    "license",
    "recipe_origin",
    "source_git_url",
    "source_url",
    "summary",
)
ABOUT_MAP_FIELDS = ("keywords", "identifiers", "tags")

MAX_CONDA_TIMESTAMP = 253402300799


def calculate_file_hashes_and_size(info, file):
    BLOCK_SIZE = 1024 * 1024  # 1MiB
    md5 = hashlib.md5()
    sha = hashlib.sha256()
    file.seek(0)
    size = 0
    while True:
        b = file.read(BLOCK_SIZE)
        if len(b) > 0:
            size += len(b)
            md5.update(b)
            sha.update(b)
        else:
            break
    info["size"] = size
    info["md5"] = md5.hexdigest()
    info["sha256"] = sha.hexdigest()


def get_subdir_compat(info):
    "compatibility layer for subdir property of packages"

    subdir = info.get("subdir", None)
    if not subdir:
        arch = info["arch"]
        platform = info["platform"]
        if arch == "x86_64":
            subdir = platform + '-64'
    return subdir


class CondaInfo:
    def __init__(self, file, filename, lazy=False):
        self.channeldata = {}
        self.package_format = None
        self._file = file
        self._filename = filename
        if filename.endswith(".conda"):
            self.package_format = db_models.PackageFormatEnum.conda
        else:
            self.package_format = db_models.PackageFormatEnum.tarbz2
        if not lazy:
            self._parse_conda()

    def __getattr__(self, name):
        # lazily extract the conda info from package files
        if name in ["info", "about", "paths", "run_exports", "files"]:
            self._parse_conda()
        return getattr(self, name)

    def _map_channeldata(self):
        channeldata = {}
        channeldata["packagename"] = self.info["name"]
        timestamp = int(self.info.get("timestamp", time.time()))
        if timestamp > MAX_CONDA_TIMESTAMP:
            # Convert timestamp from milliseconds to seconds
            timestamp //= 1000
        channeldata["timestamp"] = timestamp
        subdir = self.info.get("subdir", None)
        channeldata["subdirs"] = [subdir]

        for field in INFO_BOOLEAN_FIELDS:
            channeldata[field] = False

        for field in ABOUT_OPTIONAL_FIELDS:
            if field in self.about:
                channeldata[field] = self.about[field]
            else:
                channeldata[field] = None

        for field in INFO_OPTIONAL_FIELDS:
            if field in self.info:
                channeldata[field] = self.info[field]
            else:
                channeldata[field] = None

        for field in ABOUT_MAP_FIELDS:
            if (field in self.about) and (len(self.about[field]) > 0):
                channeldata[field] = self.about[field]
            else:
                channeldata[field] = None

        for path in self.paths.get("paths", []):
            pathname = path["_path"]
            if "/etc/conda/deactivate.d/" in pathname:
                channeldata["deactivate.d"] = True
            elif "etc/conda/activate.d/" in pathname:
                channeldata["activate.d"] = True
            if "file_mode" in path:
                if path["file_mode"] == "binary":
                    channeldata["binary_prefix"] = True
                elif path["file_mode"] == "text":
                    channeldata["text_prefix"] = True
            for linkname in ("pre-link", "post-link", "pre-unlink"):
                if fnmatch.fnmatch(pathname, f"*/.*-{linkname}.*"):
                    channeldata[linkname.replace("-", "_")] = True

        channeldata["run_exports"] = self.run_exports

        self.channeldata = channeldata

    def _load_jsons(self, tar):
        self.info = json.load(tar.extractfile("info/index.json"))

        self.info['subdir'] = get_subdir_compat(self.info)

        try:
            self.about = json.load(tar.extractfile("info/about.json"))
        except KeyError:
            self.about = {}
        try:
            self.paths = json.load(tar.extractfile("info/paths.json"))
        except KeyError:
            self.paths = {}
        with tar.extractfile("info/files") as fp:
            self.files = fp.readlines()

        try:
            exports_file = tar.extractfile("info/run_exports.json")
        except KeyError:
            self.run_exports = {}
        else:
            self.run_exports = json.load(exports_file)

        self._map_channeldata()

    def _calculate_file_hashes(self, file):
        calculate_file_hashes_and_size(self.info, file)

    def _parse_conda(self):
        file = self._file

        # workaround for https://github.com/python/cpython/pull/3249
        if not hasattr(file, "seekable"):
            file.seekable = file._file.seekable

        file.seek(0)
        filehandle = file
        if self.package_format == db_models.PackageFormatEnum.conda:
            with ZipFile(filehandle) as zf:
                infotars = [_ for _ in zf.namelist() if _.startswith("info-")]
                infotar = infotars[0]
                with zf.open(infotar) as zfobj:
                    if infotar.endswith(".zst"):
                        zstd = zstandard.ZstdDecompressor()
                        # zstandard.stream_reader cannot seek backwards
                        # and tarfile.extractfile() seeks backwards
                        fobj = BytesIO(zstd.stream_reader(zfobj).read())
                    else:
                        fobj = zfobj
                    with tarfile.open(fileobj=fobj, mode="r") as tar:
                        self._load_jsons(tar)
        else:
            try:
                with tarfile.open(fileobj=filehandle, mode="r:bz2") as tar:
                    self._load_jsons(tar)
            except tarfile.ReadError as e:
                raise PackageError(e.args[0])

        self._calculate_file_hashes(file)

        self.info = dict(sorted(self.info.items(), key=lambda item: item[0]))



import uuid
from datetime import datetime, timedelta
from pathlib import Path
from unittest.mock import ANY

import pytest

from quetz import db_models
from quetz.authorization import (
    MAINTAINER,
    MEMBER,
    OWNER,
    SERVER_MAINTAINER,
    SERVER_MEMBER,
    SERVER_OWNER,
)
from quetz.condainfo import CondaInfo
from quetz.config import Config
from quetz.jobs.models import Job
from quetz.jobs.runner import Supervisor
from quetz.testing.mockups import MockWorker


@pytest.fixture
def maintainer(user, db):
    user.role = "maintainer"
    db.commit()


@pytest.mark.parametrize(
    "user_role,expected_status",
    [("owner", 201), ("maintainer", 201), ("member", 201), (None, 403)],
)
def test_create_normal_channel_permissions(auth_client, expected_status):
    response = auth_client.post(
        "/api/channels",
        json={
            "name": "test-create-channel",
            "private": False,
        },
    )
    assert response.status_code == expected_status


@pytest.mark.parametrize("channel_role", ["owner", "maintainer", "member"])
@pytest.mark.parametrize("user_role", ["owner", "maintainer", "member", None])
def test_delete_channel_permissions(
    db, auth_client, public_channel, user_role, channel_role
):
    response = auth_client.delete(f"/api/channels/{public_channel.name}")

    channel = (
        db.query(db_models.Channel)
        .filter(db_models.Channel.name == public_channel.name)
        .one_or_none()
    )

    if user_role in ["owner", "maintainer"] or channel_role in ["owner", "maintainer"]:
        assert response.status_code == 200
        assert channel is None
    else:
        assert response.status_code == 403
        assert channel is not None


@pytest.mark.parametrize("user_role", ["owner"])
def test_delete_channel_with_packages(
    db, auth_client, private_channel, private_package_version, config: Config
):
    pkg_store = config.get_package_store()
    pkg_store.add_file("test-file", private_channel.name, "test_file.txt")
    pkg_store.add_file("second", private_channel.name, "subdir/second_file.txt")

    response = auth_client.delete(f"/api/channels/{private_channel.name}")

    channel = (
        db.query(db_models.Channel)
        .filter(db_models.Channel.name == private_channel.name)
        .one_or_none()
    )

    version = (
        db.query(db_models.PackageVersion)
        .filter_by(package_name=private_package_version.package_name)
        .one_or_none()
    )
    package = (
        db.query(db_models.Package)
        .filter_by(name=private_package_version.package_name)
        .one_or_none()
    )

    files = pkg_store.list_files(private_channel.name)

    assert response.status_code == 200
    assert channel is None
    assert version is None
    assert package is None
    assert not files


@pytest.mark.parametrize(
    "endpoint",
    [
        "/api/channels/{channel_name}",
        "/api/channels/{channel_name}/packages",
        "/api/channels/{channel_name}/packages/{package_name}",
        "/api/channels/{channel_name}/packages/{package_name}/versions",
    ],
)
@pytest.mark.parametrize(
    "user_role,expected_status",
    [("owner", 200), ("maintainer", 200), ("member", 403), (None, 403)],
)
def test_permissions_channel_endpoints(
    auth_client,
    private_channel,
    expected_status,
    endpoint,
    private_package,
    private_package_version,
):
    response = auth_client.get(
        endpoint.format(
            channel_name=private_channel.name, package_name=private_package.name
        )
    )
    assert response.status_code == expected_status


@pytest.fixture
def sync_supervisor(db, dao, config):
    "supervisor with synchronous test worker"
    manager = MockWorker(config, db, dao)
    supervisor = Supervisor(db, manager)
    return supervisor


@pytest.mark.parametrize(
    "action", ['reindex', 'generate_indexes', 'synchronize_metrics']
)
@pytest.mark.parametrize(
    "channel_role,expected_code",
    [("owner", 200), ("maintainer", 200), ("member", 403), (None, 403)],
)
def test_channel_action_reindex(
    auth_client,
    public_channel,
    expected_code,
    action,
    user,
    remove_jobs,
    sync_supervisor,
):
    response = auth_client.put(
        f"/api/channels/{public_channel.name}/actions", json={"action": action}
    )

    assert response.status_code == expected_code
    if expected_code == 200:
        job_data = response.json()
        assert job_data == {
            "created": ANY,
            "id": ANY,
            "items_spec": None,
            "manifest": action,
            "owner_id": str(uuid.UUID(bytes=user.id)),
            "status": "pending",
            "repeat_every_seconds": None,
            "start_at": None,
        }
        job_id = job_data['id']
    else:
        return

    response = auth_client.get(f"/api/jobs/{job_id}")
    assert response.status_code == 200
    job_data = response.json()
    assert job_data == {
        "created": ANY,
        "id": job_id,
        "items_spec": None,
        "manifest": action,
        "owner_id": str(uuid.UUID(bytes=user.id)),
        "status": "pending",
        "repeat_every_seconds": None,
        "start_at": None,
    }

    sync_supervisor.run_jobs()

    response = auth_client.get(
        f"/api/jobs/{job_id}/tasks?"
        "status=created&status=running&status=pending&status=success"
    )

    assert response.status_code == 200
    all_tasks = response.json()['result']

    assert all_tasks == [
        {
            "job_id": job_id,
            'created': ANY,
            'id': ANY,
            'package_version': {},
            'status': 'created',
        }
    ]


@pytest.mark.parametrize("wait_seconds", [None, -5, 5, 0])
def test_create_delayed_action(
    auth_client,
    public_channel,
    user,
    remove_jobs,
    wait_seconds,
    sync_supervisor,
):
    action = "reindex"
    now = datetime.utcnow()
    if wait_seconds is None:
        start_at = None
    else:
        delta = timedelta(seconds=wait_seconds)
        start_at = (now + delta).isoformat()

    response = auth_client.put(
        f"/api/channels/{public_channel.name}/actions",
        json={"action": action, "start_at": start_at},
    )

    assert response.status_code == 200
    job_data = response.json()
    job_id = job_data['id']

    sync_supervisor.run_once()

    response = auth_client.get(
        f"/api/jobs/{job_id}/tasks?"
        "status=created&status=running&status=pending&status=success"
    )

    assert response.status_code == 200
    all_tasks = response.json()['result']

    if wait_seconds is not None and wait_seconds > 0:
        assert not all_tasks
    else:
        assert len(all_tasks) == 1


@pytest.mark.parametrize(
    "channel_role,expected_code",
    [("owner", 200), ("maintainer", 200), ("member", 403), (None, 403)],
)
def test_get_channel_members(auth_client, public_channel, expected_code):
    response = auth_client.get(f"/api/channels/{public_channel.name}/members")

    assert response.status_code == expected_code


@pytest.fixture
def remove_package_versions(db):
    yield
    db.query(db_models.PackageVersion).delete()


@pytest.fixture
def remove_jobs(db):
    yield
    db.query(Job).delete()


def test_channel_names_are_case_insensitive(
    auth_client, maintainer, remove_package_versions
):
    channel_name = "MyChanneL"

    response = auth_client.post(
        "/api/channels", json={"name": channel_name, "private": False}
    )

    assert response.status_code == 201

    response = auth_client.get(f"/api/channels/{channel_name}")

    assert response.status_code == 200
    assert response.json()["name"] == channel_name

    response = auth_client.get(f"/api/channels/{channel_name.lower()}")

    assert response.status_code == 200
    assert response.json()["name"] == channel_name

    response = auth_client.get(f"/api/channels/{channel_name.lower()}/packages")

    assert response.status_code == 200

    assert response.json() == []
    package_filename = "test-package-0.1-0.tar.bz2"
    with open(package_filename, "rb") as fid:
        files = {"files": (package_filename, fid)}
        response = auth_client.post(f"/api/channels/{channel_name}/files/", files=files)

    response = auth_client.get(f"/api/channels/{channel_name.lower()}/packages")
    assert response.status_code == 200
    assert len(response.json()) == 1

    response = auth_client.get(f"/api/channels/{channel_name}/packages")
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert response.json()[0]["name"] == "test-package"

    response = auth_client.get(
        f"/api/channels/{channel_name}/packages/test-package/versions"
    )
    assert response.status_code == 200
    assert len(response.json()) == 1

    response = auth_client.get(
        f"/api/channels/{channel_name.lower()}/packages/test-package/versions"
    )
    assert response.status_code == 200
    assert len(response.json()) == 1

    response = auth_client.get(f"/get/{channel_name.lower()}/linux-64/repodata.json")
    assert response.status_code == 200
    assert package_filename in response.json()["packages"]

    response = auth_client.get(f"/get/{channel_name.lower()}")
    assert response.status_code == 200

    response = auth_client.get(f"/get/{channel_name}")
    assert response.status_code == 200

    response = auth_client.get(f"/get/{channel_name}/linux-64/repodata.json")
    assert response.status_code == 200
    assert package_filename in response.json()["packages"]

    response = auth_client.get(
        f"/get/{channel_name.lower()}/linux-64/{package_filename}"
    )
    assert response.status_code == 200

    response = auth_client.get(f"/get/{channel_name}/linux-64/{package_filename}")
    assert response.status_code == 200


def test_unique_channel_names_are_case_insensitive(auth_client, maintainer):
    channel_name = "MyChanneL"

    response = auth_client.post(
        "/api/channels", json={"name": channel_name, "private": False}
    )

    assert response.status_code == 201

    response = auth_client.post(
        "/api/channels", json={"name": channel_name.lower(), "private": False}
    )

    assert response.status_code == 409
    assert f"{channel_name.lower()} exists" in response.json()["detail"]

    response = auth_client.post(
        "/api/channels", json={"name": channel_name.upper(), "private": False}
    )

    assert response.status_code == 409
    assert f"{channel_name.upper()} exists" in response.json()["detail"]


def test_unicode_channel_names(auth_client, maintainer):
    channel_name = "검은맘바"

    response = auth_client.post(
        "/api/channels", json={"name": channel_name, "private": False}
    )

    assert response.status_code == 422
    assert (
        response.json()["detail"]
        == "only ASCII characters should be used in channel name"
    )

    response = auth_client.get("/api/channels")

    assert response.status_code == 200
    assert len(response.json()) == 0

    response = auth_client.get(f"/api/channels/{channel_name}")

    assert response.status_code == 404


def test_accents_make_unique_channel_names(auth_client, maintainer):
    channel_names = ["żmija", "zmija", "grün", "grun"]

    for i, name in enumerate(channel_names):
        response = auth_client.post(
            "/api/channels", json={"name": name, "private": False}
        )
        if (i % 2) == 0:
            assert response.status_code == 422
            assert (
                response.json()["detail"]
                == "only ASCII characters should be used in channel name"
            )
        else:
            assert response.status_code == 201

    response = auth_client.get("/api/channels")

    assert response.status_code == 200

    assert len(response.json()) == 2


def test_upload_package_version_to_channel(
    auth_client,
    public_channel,
    maintainer,
    db,
    config,
    remove_package_versions,
):
    pkgstore = config.get_package_store()

    assert public_channel.size == 0

    package_filename = "test-package-0.1-0.tar.bz2"
    with open(package_filename, "rb") as fid:
        files = {"files": (package_filename, fid)}
        response = auth_client.post(
            f"/api/channels/{public_channel.name}/files/",
            files=files,
        )

    with open(package_filename, "rb") as fid:
        condainfo = CondaInfo(fid, package_filename)
        condainfo._parse_conda()

    assert response.status_code == 201
    db.refresh(public_channel)
    assert public_channel.size == condainfo.info["size"]
    assert pkgstore.serve_path(
        public_channel.name, str(Path(condainfo.info["subdir"]) / package_filename)
    )


@pytest.mark.parametrize("user_role", ["maintainer"])
@pytest.mark.parametrize(
    "config_extra,expected_size_limit",
    [("", None), ("[quotas]\nchannel_quota=100\n", 100), ("[quotas]\n", None)],
)
def test_create_channel_default_quotas(auth_client, expected_size_limit, db, config):
    name = "test-create-channel-with-quotas"
    response = auth_client.post(
        "/api/channels",
        json={
            "name": name,
            "private": False,
        },
    )

    assert response.status_code == 201

    channel = db.query(db_models.Channel).filter(db_models.Channel.name == name).one()

    if expected_size_limit is None:
        assert channel.size_limit is None
    else:
        assert channel.size_limit == expected_size_limit


@pytest.mark.parametrize("user_role", [SERVER_MAINTAINER, SERVER_OWNER, SERVER_MEMBER])
@pytest.mark.parametrize(
    "config_extra",
    [
        pytest.param("", id="no-config"),
        pytest.param("[quotas]\nchannel_quota=10\n", id="with-config"),
    ],
)
def test_create_channel_with_limits(auth_client, db, user_role):
    name = "test-create-channel-with-quotas"
    response = auth_client.post(
        "/api/channels",
        json={"name": name, "private": False, "size_limit": 101},
    )

    channel = (
        db.query(db_models.Channel).filter(db_models.Channel.name == name).one_or_none()
    )
    if user_role in [SERVER_MAINTAINER, SERVER_OWNER]:
        assert response.status_code == 201
        assert channel.size_limit == 101
    else:
        assert response.status_code == 403
        assert channel is None
        assert "set channel size limit" in response.json()["detail"][0]


@pytest.mark.parametrize("user_role", [SERVER_MAINTAINER, SERVER_OWNER])
def test_set_channel_size_limit(auth_client, db, public_channel):
    assert public_channel.size_limit is None

    response = auth_client.patch(
        f"/api/channels/{public_channel.name}",
        json={"size_limit": 101},
    )

    assert response.status_code == 200

    data = response.json()
    assert data["name"] == public_channel.name
    assert data["size_limit"] == 101

    db.refresh(public_channel)

    assert public_channel.size_limit == 101


@pytest.mark.parametrize("ttl", [None, 1200, 36000])
@pytest.mark.parametrize("user_role", [SERVER_OWNER])
def test_create_channel_ttl(auth_client, db, user_role, ttl):
    name = "test-create-channel-ttl"
    payload = {"name": name, "private": False}
    if ttl is not None:
        payload["ttl"] = ttl
    response = auth_client.post("/api/channels", json=payload)
    channel = (
        db.query(db_models.Channel).filter(db_models.Channel.name == name).one_or_none()
    )
    assert response.status_code == 201
    if ttl is None:
        expected_ttl = 36000
    else:
        expected_ttl = ttl
    assert channel.ttl == expected_ttl


@pytest.mark.parametrize("user_role", [SERVER_OWNER])
def test_update_channel_ttl(auth_client, db, public_channel):
    assert public_channel.ttl == 36000

    response = auth_client.patch(
        f"/api/channels/{public_channel.name}",
        json={"ttl": 1200},
    )

    assert response.status_code == 200

    data = response.json()
    assert data["name"] == public_channel.name
    assert data["ttl"] == 1200

    db.refresh(public_channel)

    assert public_channel.ttl == 1200


@pytest.mark.parametrize("user_role", [SERVER_OWNER])
@pytest.mark.parametrize(
    "attr_dict",
    [
        {"name": "new-name"},
        {"mirror_channel_url": "http://test", "mirror_mode": "proxy"},
    ],
)
def test_update_channel_forbidden_attributes(
    auth_client,
    db,
    public_channel,
    attr_dict,
):
    response = auth_client.patch(
        f"/api/channels/{public_channel.name}",
        json=attr_dict,
    )

    assert response.status_code == 422
    assert "can not be changed" in response.json()["detail"]


@pytest.mark.parametrize("user_role", [SERVER_OWNER])
@pytest.mark.parametrize("name,value", [("private", True)])
def test_update_channel_attributes(auth_client, db, public_channel, name, value):
    response = auth_client.patch(
        f"/api/channels/{public_channel.name}",
        json={name: value},
    )

    assert response.status_code == 200

    data = response.json()
    assert data["name"] == public_channel.name
    assert data[name] == value

    db.refresh(public_channel)

    assert getattr(public_channel, name) == value


@pytest.mark.parametrize(
    "user_role", [SERVER_OWNER, SERVER_MEMBER, SERVER_MAINTAINER, None]
)
@pytest.mark.parametrize("channel_role", [OWNER, MAINTAINER, MEMBER])
def test_update_channel_permissions(
    auth_client, db, public_channel, user_role, channel_role
):
    response = auth_client.patch(
        f"/api/channels/{public_channel.name}",
        json={"private": False},
    )

    if user_role in [SERVER_OWNER, SERVER_MAINTAINER]:
        assert response.status_code == 200
    elif channel_role in [OWNER, MAINTAINER]:
        assert response.status_code == 200
    else:
        assert response.status_code == 403


@pytest.fixture
def remote_session(app, request, public_channel, auth_client):
    mirror_url = f"{auth_client.base_url}/get/{public_channel.name}"

    from quetz.main import get_remote_session

    class dummy_response:
        status_code = 200

        def json(self):
            return {"mirror_channel_url": mirror_url}

    class dummy_session:
        def get(self, url):
            return dummy_response()

    app.dependency_overrides[get_remote_session] = dummy_session

    return dummy_session


def test_register_mirror(auth_client, public_channel, db, remote_session):
    mirror_url = "http://mirror_url/get/my-channel"

    response = auth_client.post(
        f"/api/channels/{public_channel.name}/mirrors", json={"url": mirror_url}
    )

    m = db.query(db_models.ChannelMirror).filter_by(url=mirror_url).one_or_none()

    assert response.status_code == 201
    assert m
    assert m.last_synchronised is None

    response = auth_client.get(f"/api/channels/{public_channel.name}/mirrors")

    assert response.status_code == 200

    assert response.json() == [
        {
            "url": mirror_url,
            "id": ANY,
            "api_endpoint": "http://mirror_url/api/channels/my-channel",
            "metrics_endpoint": "http://mirror_url/metrics/channels/my-channel",
        }
    ]

    mirror_id = response.json()[0]["id"]
    response = auth_client.delete(
        f"/api/channels/{public_channel.name}/mirrors/{mirror_id}"
    )

    assert response.status_code == 200

    response = auth_client.get(f"/api/channels/{public_channel.name}/mirrors")

    assert response.status_code == 200

    assert response.json() == []


def test_url_with_slash(auth_client, public_channel, db, remote_session):
    mirror_url = "http://mirror_url"

    response = auth_client.post(
        f"/api/channels/{public_channel.name}/mirrors/",
        json={"url": mirror_url},
        follow_redirects=False,
    )

    assert response.status_code == 307


@pytest.mark.parametrize(
    "auth_user,owned_channels",
    [("bartosz", ["my-channel"]), ("other", ["private-channel"])],
)
@pytest.mark.parametrize("include_public", [1, 0])
@pytest.mark.parametrize("endpoint", ["channels", "paginated/channels"])
def test_list_channels(
    client,
    private_channel,
    public_channel,
    auth_user,
    include_public,
    owned_channels,
    endpoint,
):
    response = client.get(f"/api/dummylogin/{auth_user}")
    assert response.status_code == 200

    response = client.get(f"/api/{endpoint}?public={include_public}")
    assert response.status_code == 200
    results = response.json()

    try:
        results = results['result']
    except TypeError:
        # non-paginated response
        pass

    channel_names = {r["name"] for r in results}

    if include_public:
        expected_channels = {public_channel.name, *owned_channels}
    else:
        expected_channels = set(owned_channels)

    assert channel_names == expected_channels


@pytest.mark.parametrize("endpoint", ["/api/channels/{channel_name}", "/api/channels"])
def test_channel_package_members_count(
    auth_client, public_channel, db, private_channel, other_user, endpoint
):
    response = auth_client.get(endpoint.format(channel_name=public_channel.name))
    assert response.status_code == 200

    channel_data = response.json()
    if isinstance(channel_data, list):
        channel_data = channel_data[0]

    assert channel_data['members_count'] == 1
    assert channel_data['packages_count'] == 0

    package = db_models.Package(channel=public_channel, name="test-package")
    db.add(package)
    db.commit()

    response = auth_client.get(endpoint.format(channel_name=public_channel.name))
    channel_data = response.json()
    if isinstance(channel_data, list):
        channel_data = channel_data[0]

    assert channel_data['packages_count'] == 1
    assert channel_data['members_count'] == 1

    package = db_models.Package(
        channel=private_channel, name="test-package-different-channel"
    )
    db.add(package)
    db.commit()

    response = auth_client.get(endpoint.format(channel_name=public_channel.name))
    channel_data = response.json()
    if isinstance(channel_data, list):
        channel_data = channel_data[0]

    assert channel_data['packages_count'] == 1
    assert channel_data['members_count'] == 1

    channel_member = db_models.ChannelMember(
        channel=public_channel, user=other_user, role="member"
    )
    db.add(channel_member)
    db.commit()

    response = auth_client.get(endpoint.format(channel_name=public_channel.name))
    channel_data = response.json()
    if isinstance(channel_data, list):
        channel_data = channel_data[0]

    assert channel_data['packages_count'] == 1
    assert channel_data['members_count'] == 2



import json

import pytest

from quetz.db_models import PackageVersion
from quetz.pkgstores import PackageStore
from quetz.rest_models import Channel
from quetz.tasks.indexing import validate_packages
from quetz.tasks.reindexing import reindex_packages_from_store


@pytest.fixture
def pkgstore(config):
    pkgstore = config.get_package_store()
    return pkgstore


@pytest.fixture
def package_filenames():
    return ["test-package-0.1-0.tar.bz2", "test-package-0.2-0.tar.bz2"]


@pytest.fixture
def package_files(pkgstore: PackageStore, channel_name, package_filenames):
    pkgstore.create_channel(channel_name)
    for filename in package_filenames:
        with open(filename, 'rb') as fid:
            content = fid.read()
        pkgstore.add_file(content, channel_name, f"linux-64/{filename}")


@pytest.fixture
def channel(dao, user, channel_name):
    channel_data = Channel(name=channel_name, private=False)
    channel = dao.create_channel(channel_data, user.id, "owner")

    return channel


@pytest.fixture
def remove_package_versions(db):
    # clean up the created package versions after the test

    yield

    db.query(PackageVersion).delete()
    db.commit()


def test_reindex_package_files(
    config,
    user,
    package_files,
    channel,
    db,
    dao,
    pkgstore: PackageStore,
    package_filenames,
    remove_package_versions,
):
    user_id = user.id
    reindex_packages_from_store(dao, config, channel.name, user_id)
    db.refresh(channel)

    assert channel.packages
    assert channel.packages[0].name == "test-package"
    assert channel.packages[0].members[0].user.username == user.username
    assert (
        channel.packages[0].package_versions[0].version == '0.1'
        or channel.packages[0].package_versions[0].version == '0.2'
    )
    assert (
        channel.packages[0].package_versions[1].version == '0.1'
        or channel.packages[0].package_versions[1].version == '0.2'
    )

    repodata = pkgstore.serve_path(channel.name, "linux-64/repodata.json")
    repodata = json.load(repodata)
    assert repodata
    assert len(repodata['packages']) == 2
    assert set(repodata["packages"].keys()) == set(package_filenames)


def test_validate_packages(
    config,
    user,
    package_files,
    channel,
    channel_name,
    db,
    dao,
    pkgstore: PackageStore,
    package_filenames,
):
    user_id = user.id
    reindex_packages_from_store(dao, config, channel.name, user_id)
    db.refresh(channel)

    assert channel.packages
    assert channel.packages[0].name == "test-package"
    assert channel.packages[0].members[0].user.username == user.username
    assert (
        channel.packages[0].package_versions[0].version == '0.1'
        or channel.packages[0].package_versions[0].version == '0.2'
    )
    assert (
        channel.packages[0].package_versions[1].version == '0.1'
        or channel.packages[0].package_versions[1].version == '0.2'
    )

    repodata = pkgstore.serve_path(channel.name, "linux-64/repodata.json")
    repodata = json.load(repodata)
    assert repodata
    assert len(repodata['packages']) == 2
    assert set(repodata["packages"].keys()) == set(package_filenames)

    remaining_pkg = channel.packages[0].package_versions[1].filename
    pkgstore.delete_file(
        channel.name, f'linux-64/{channel.packages[0].package_versions[0].filename}'
    )

    validate_packages(dao, pkgstore, channel_name)

    db.refresh(channel)

    repodata = pkgstore.serve_path(channel.name, "linux-64/repodata.json")
    repodata = json.load(repodata)
    assert repodata
    assert len(repodata['packages']) == 1
    assert set(repodata["packages"].keys()) == set([remaining_pkg])
    assert len(channel.packages[0].package_versions) == 1

    pkgstore.add_file(b"wrong_size", channel_name, f"linux-64/{remaining_pkg}")

    validate_packages(dao, pkgstore, channel_name)

    db.refresh(channel)

    repodata = pkgstore.serve_path(channel.name, "linux-64/repodata.json")
    repodata = json.load(repodata)
    assert repodata
    assert len(repodata['packages']) == 0
    assert set(repodata["packages"].keys()) == set([])
    assert len(channel.packages[0].package_versions) == 0



import uuid

import pytest

from quetz.db_models import ApiKey, ChannelMember, PackageMember, User


def test_validate_user_role_names(user, client, other_user, db):
    # test validation of role names

    response = client.put("/api/users/bartosz/role", json={"role": "UNDEFINED"})

    assert response.status_code == 422
    assert "member" in response.json()["detail"][0]["msg"]


@pytest.mark.parametrize(
    "target_user,user_role,target_user_role,expected_status",
    [
        ("other", None, "member", 403),
        ("other", "member", "member", 403),
        ("other", "member", "maintainer", 403),
        ("other", "member", "owner", 403),
        ("other", "maintainer", "member", 200),
        ("other", "maintainer", "maintainer", 403),
        ("other", "maintainer", "owner", 403),
        ("other", "owner", "member", 200),
        ("other", "owner", "maintainer", 200),
        ("other", "owner", "owner", 200),
        ("other", "owner", None, 200),
        ("missing_user", "owner", "member", 404),
    ],
)
def test_set_user_role(
    auth_client,
    other_user,
    target_user,
    user_role,
    target_user_role,
    expected_status,
):
    # test changing role

    response = auth_client.put(
        f"/api/users/{target_user}/role", json={"role": target_user_role}
    )
    assert response.status_code == expected_status

    # test if role assigned if previous request was successful
    if response.status_code == 200:
        get_response = auth_client.get(f"/api/users/{target_user}/role")

        assert get_response.status_code == 200
        assert get_response.json()["role"] == target_user_role


@pytest.mark.parametrize(
    "target_user,user_role,expected_status",
    [
        ("other", None, 403),
        ("other", "member", 403),
        ("other", "maintainer", 200),
        ("other", "owner", 200),
        ("bartosz", None, 200),
        ("bartosz", "member", 200),
        ("bartosz", "maintainer", 200),
        ("bartosz", "owner", 200),
    ],
)
def test_get_user_role(auth_client, other_user, target_user, expected_status, db):
    # test reading the role

    response = auth_client.get(f"/api/users/{target_user}/role")
    assert response.status_code == expected_status

    expected_role = db.query(User).filter(User.username == target_user).first().role

    if response.status_code == 200:
        assert response.json()["role"] == expected_role


def test_get_set_user_role_without_login(user, client):
    # test permissions without logged-in user

    response = client.get("/api/users/bartosz/role")

    assert response.status_code == 401

    response = client.put("/api/users/bartosz/role", json={"role": "member"})

    assert response.status_code == 401


@pytest.mark.parametrize(
    "user_role,target_user,expected_status",
    [
        ("owner", "other", 200),
        ("maintainer", "other", 200),
        ("member", "other", 403),
        (None, "other", 403),
        ("owner", "bartosz", 200),
        ("maintainer", "bartosz", 200),
        ("member", "bartosz", 200),
        (None, "bartosz", 200),
    ],
)
def test_get_user_permissions(
    other_user,
    target_user,
    auth_client,
    expected_status,
):
    response = auth_client.get(f"/api/users/{target_user}")

    assert response.status_code == expected_status

    if response.status_code == 200:
        assert response.json()["username"] == target_user


@pytest.mark.parametrize("paginated", [False, True])
@pytest.mark.parametrize(
    "user_role,query,expected_n_users",
    [
        ("owner", "", 2),
        ("maintainer", "", 2),
        ("member", "", 1),
        (None, "", 1),
        ("owner", "bar", 1),
        (None, "bar", 1),
        ("owner", "oth", 1),
        (None, "oth", 0),
    ],
)
def test_get_users_permissions(
    other_user, auth_client, expected_n_users, query, paginated
):
    if paginated:
        response = auth_client.get(f"/api/paginated/users?q={query}")
        user_list = response.json()["result"]
    else:
        response = auth_client.get(f"/api/users?q={query}")
        user_list = response.json()

    assert response.status_code == 200

    assert len(user_list) == expected_n_users


@pytest.fixture
def api_keys(user, other_user, db):
    users = [user, other_user]

    for key_user in users:
        for key_owner in users:
            key = ApiKey(key=str(uuid.uuid4()), user=key_user, owner=key_owner)
            db.add(key)

    db.commit()


@pytest.mark.parametrize(
    "user_role,target_user,expected_status",
    [
        ("owner", "other", 200),
        ("maintainer", "other", 200),
        ("member", "other", 403),
        (None, "other", 403),
        # user can always remove oneself
        ("maintainer", "bartosz", 200),
        ("member", "bartosz", 200),
        (None, "bartosz", 200),
    ],
)
def test_delete_user_permission(
    other_user, auth_client, db, user_role, target_user, expected_status, user, api_keys
):
    response = auth_client.delete(f"/api/users/{target_user}")

    deleted_user = db.query(User).filter(User.username == target_user).one_or_none()

    if expected_status == 200:
        assert response.status_code == 200
        assert deleted_user
        assert not deleted_user.profile
        assert not deleted_user.identities
        assert not deleted_user.api_keys_owner
        assert not deleted_user.api_keys_user

    else:
        assert response.status_code == expected_status
        assert deleted_user
        assert deleted_user.profile
        assert deleted_user.identities
        assert deleted_user.api_keys_owner
        assert deleted_user.api_keys_user

    # check if other users were not accidently removed
    existing_user = db.query(User).filter(User.username != target_user).one_or_none()
    assert existing_user
    assert existing_user.profile
    assert existing_user.api_keys_owner
    assert existing_user.api_keys_user


@pytest.mark.parametrize("user_role", ["owner"])
def test_get_users_without_profile(auth_client, other_user_without_profile, user):
    response = auth_client.get("/api/users")

    assert response.status_code == 200
    users = response.json()
    assert len(users) == 1
    assert users[0]["username"] == user.username

    response = auth_client.get(f"/api/users/{other_user_without_profile.username}")

    assert response.status_code == 404


@pytest.mark.parametrize(
    "http_user,expected_role", [("bartosz", "member"), ("other", "owner")]
)
def test_list_user_channels(
    user, client, other_user, db, private_channel, http_user, expected_role
):
    member = ChannelMember(
        channel_name=private_channel.name, user_id=user.id, role="member"
    )
    db.add(member)
    db.commit()

    response = client.get(f"/api/dummylogin/{http_user}")
    assert response.status_code == 200

    response = client.get(f"/api/users/{http_user}/channels")
    assert response.status_code == 200

    assert response.json() == [{"name": private_channel.name, "role": expected_role}]

    response = client.get(f"/api/paginated/users/{http_user}/channels")
    assert response.status_code == 200

    assert response.json()["result"] == [
        {"name": private_channel.name, "role": expected_role}
    ]


@pytest.mark.parametrize(
    "http_user,expected_role", [("bartosz", "member"), ("other", "owner")]
)
@pytest.mark.parametrize(
    "endpoint",
    [
        "/api/users/{http_user}/packages",
        "/api/paginated/users/{http_user}/packages",
    ],
)
def test_list_user_packages(
    user,
    client,
    other_user,
    db,
    private_channel,
    private_package,
    http_user,
    expected_role,
    endpoint,
):
    member = PackageMember(
        channel=private_channel, package=private_package, user=user, role="member"
    )
    db.add(member)
    db.commit()

    response = client.get(f"/api/dummylogin/{http_user}")
    assert response.status_code == 200

    response = client.get(endpoint.format(http_user=http_user))
    assert response.status_code == 200

    data = response.json()

    if "result" in data:
        data = data["result"]

    assert data == [
        {
            "name": private_package.name,
            "channel_name": private_channel.name,
            "role": expected_role,
        }
    ]



from fastapi.testclient import TestClient


def test_get_health_ready(client: TestClient):
    response = client.get("/health/ready")
    assert response.status_code == 200


def test_get_health_live(client: TestClient):
    response = client.get("/health/live")
    assert response.status_code == 200



import logging
from datetime import datetime
from typing import Optional

from fastapi import HTTPException, status

from quetz import authorization, dao, db_models
from quetz.jobs.dao import JobsDao
from quetz.rest_models import ChannelActionEnum

from . import assertions

logger = logging.getLogger("quetz")


def assert_channel_action(action, channel):
    if action == ChannelActionEnum.synchronize:
        action_allowed = assertions.can_channel_synchronize(channel)
    elif action == ChannelActionEnum.synchronize_repodata:
        action_allowed = assertions.can_channel_synchronize(channel)
    elif action == ChannelActionEnum.validate_packages:
        action_allowed = assertions.can_channel_validate_package_cache(channel)
    elif action == ChannelActionEnum.generate_indexes:
        action_allowed = assertions.can_channel_reindex(channel)
    elif action == ChannelActionEnum.reindex:
        action_allowed = assertions.can_channel_reindex(channel)
    elif action == ChannelActionEnum.synchronize_metrics:
        action_allowed = assertions.can_channel_synchronize_metrics(channel)
    elif action == ChannelActionEnum.cleanup:
        action_allowed = assertions.can_cleanup(channel)
    elif action == ChannelActionEnum.cleanup_dry_run:
        action_allowed = assertions.can_cleanup(channel)
    else:
        action_allowed = False

    if not action_allowed:
        raise HTTPException(
            status_code=status.HTTP_405_METHOD_NOT_ALLOWED,
            detail=f"Action {action} not allowed for channel {channel.name}",
        )


class Task:
    def __init__(
        self,
        auth: authorization.Rules,
        db,
    ):
        from quetz.deps import get_config

        self.auth = auth
        self.db = db
        self.jobs_dao = JobsDao(db)
        self.dao = dao.Dao(db)
        self.pkgstore = get_config().get_package_store()

    def execute_channel_action(
        self,
        action: str,
        channel: db_models.Channel,
        start_at: Optional[datetime] = None,
        repeat_every_seconds: Optional[int] = None,
    ):
        auth = self.auth

        channel_name = channel.name
        channel_metadata = channel.load_channel_metadata()
        assert_channel_action(action, channel)

        user_id = auth.assert_user()

        if action == ChannelActionEnum.synchronize:
            auth.assert_synchronize_mirror(channel_name)
            extra_args = dict(
                channel_name=channel_name,
                includelist=channel_metadata.get('includelist', None),
                excludelist=channel_metadata.get('excludelist', None),
            )
            task = self.jobs_dao.create_job(
                action.encode('ascii'),
                user_id,
                extra_args=extra_args,
                start_at=start_at,
                repeat_every_seconds=repeat_every_seconds,
            )
        elif action == ChannelActionEnum.synchronize_repodata:
            auth.assert_synchronize_mirror(channel_name)
            extra_args = dict(
                channel_name=channel_name,
                use_repodata=True,
                includelist=channel_metadata.get('includelist', None),
                excludelist=channel_metadata.get('excludelist', None),
            )
            task = self.jobs_dao.create_job(
                action.encode('ascii'),
                user_id,
                extra_args=extra_args,
                start_at=start_at,
                repeat_every_seconds=repeat_every_seconds,
            )
        elif action == ChannelActionEnum.validate_packages:
            auth.assert_validate_package_cache(channel_name)
            extra_args = dict(channel_name=channel.name)
            task = self.jobs_dao.create_job(
                action.encode('ascii'),
                user_id,
                extra_args=extra_args,
                start_at=start_at,
                repeat_every_seconds=repeat_every_seconds,
            )
        elif action == ChannelActionEnum.generate_indexes:
            auth.assert_reindex_channel(channel_name)
            extra_args = dict(channel_name=channel.name)
            task = self.jobs_dao.create_job(
                action.encode('ascii'),
                user_id,
                extra_args=extra_args,
                start_at=start_at,
                repeat_every_seconds=repeat_every_seconds,
            )
        elif action == ChannelActionEnum.reindex:
            auth.assert_reindex_channel(channel_name)
            extra_args = dict(
                channel_name=channel_name,
            )
            task = self.jobs_dao.create_job(
                action.encode('ascii'),
                user_id,
                extra_args=extra_args,
                start_at=start_at,
                repeat_every_seconds=repeat_every_seconds,
            )
        elif action == ChannelActionEnum.synchronize_metrics:
            auth.assert_reindex_channel(channel_name)
            extra_args = dict(channel_name=channel.name)
            task = self.jobs_dao.create_job(
                action.encode('ascii'),
                user_id,
                extra_args=extra_args,
                start_at=start_at,
                repeat_every_seconds=repeat_every_seconds,
            )
        elif action in [ChannelActionEnum.cleanup, ChannelActionEnum.cleanup_dry_run]:
            auth.assert_channel_db_cleanup(channel_name)
            dry_run = action == ChannelActionEnum.cleanup_dry_run
            extra_args = dict(
                channel_name=channel_name,
                dry_run=dry_run,
            )
            task = self.jobs_dao.create_job(
                f"db_{action}".encode('ascii'),
                user_id,
                extra_args=extra_args,
                start_at=start_at,
                repeat_every_seconds=repeat_every_seconds,
            )
            task = self.jobs_dao.create_job(
                f"pkgstore_{action}".encode('ascii'),
                user_id,
                extra_args=extra_args,
                start_at=start_at,
                repeat_every_seconds=repeat_every_seconds,
            )
        else:
            raise HTTPException(
                status_code=status.HTTP_501_NOT_IMPLEMENTED,
                detail=(
                    f"Action {action} on channel {channel.name} is not implemented"
                ),
            )
        return task



import json

from .oauth2 import OAuthAuthenticator


class AzureADAuthenticator(OAuthAuthenticator):
    """Use Microsoft Azure Active Directory account to authenticate users with Quetz.

    To enable add the following to the configuration file:

    .. code::

      [azuread]
      client_id = "some-client-id-value"
      client_secret = "a-client-secret"
      tenant_id = "tenant-name or id"

    You can obtain ``client_id`` and ``client_secret`` by registering your
    application with AzureAD platfrom at this URL:
    `https://portal.azure.com/#blade/Microsoft_AAD_IAM/ActiveDirectoryMenuBlade/RegisteredApps`

    The ``tenant_id`` can either be a specific tenant's GUID identifier or one of three
    values:
    ``common``
    ``organizations``
    ``consumers``
    """

    provider = "azuread"
    collect_emails = False

    scope = "openid profile email"

    async def userinfo(self, request, token):
        response = await self.client.get(
            "https://graph.microsoft.com/oidc/userinfo", token=token
        )
        profile = response.json()

        azuread_profile = {
            "id": profile["sub"],
            "name": profile["name"],
            # TODO: the `picture` url requires to be requested with a bearer token.
            #  Thats not supported right now.
            # "avatar_url": profile["picture"],
            "avatar_url": "",
            "login": profile["email"],
        }

        if self.collect_emails:
            azuread_profile["emails"] = [
                {
                    "email": profile["email"],
                    "primary": True,
                    # AzureAD emails can always be considered to be verified.
                    # https://stackoverflow.com/questions/40618210/is-an-email-verified-in-azuread
                    "verified": True,
                }
            ]

        return azuread_profile

    async def authenticate(self, request, data=None, dao=None, config=None):
        token = await self.client.authorize_access_token(request)
        profile = await self.userinfo(request, token)

        username = profile["login"]

        # We have to remove the 'id_token' entry so we can save the token in a cookie
        # session
        token.pop('id_token', None)

        auth_state = {"token": json.dumps(token), "provider": self.provider}

        return {"username": username, "profile": profile, "auth_state": auth_state}

    def configure(self, config):
        if config.configured_section("azuread"):
            self.client_id = config.azuread_client_id
            self.client_secret = config.azuread_client_secret

            tenant_id = config.azuread_tenant_id
            self.server_metadata_url = (
                f'https://login.microsoftonline.com/{tenant_id}/v2.0/'
                f'.well-known/openid-configuration'
            )

            # oauth client params
            self.api_base_url = f'https://login.microsoftonline.com/{tenant_id}/v2.0'
            self.access_token_url = (
                f'https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token'
            )
            self.authorize_url = (
                f'https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/authorize'
            )
            # endpoints
            self.revoke_url = (
                f'https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/logout'
            )
            self.validate_token_url = 'https://graph.microsoft.com/oidc/userinfo'

            if config.configured_section("users"):
                self.collect_emails = config.users_collect_emails

            self.is_enabled = True
        else:
            self.is_enabled = False

        # call the configure of base class to set default_channel and default role
        super().configure(config)



import hashlib
import json
import os
import shutil
import time
from pathlib import Path
from typing import BinaryIO, Callable, Tuple, Union

import pytest

from quetz import hookimpl
from quetz.authorization import MAINTAINER, MEMBER, OWNER, SERVER_OWNER
from quetz.condainfo import CondaInfo
from quetz.config import Config
from quetz.dao import Dao
from quetz.db_models import ChannelMember, Package, PackageMember, PackageVersion
from quetz.errors import ValidationError
from quetz.rest_models import BaseApiKey, Channel
from quetz.tasks.indexing import update_indexes


@pytest.mark.parametrize("package_role", [OWNER, MAINTAINER, MEMBER])
@pytest.mark.parametrize("channel_role", [MEMBER])
def test_delete_package(
    auth_client, public_package, public_channel, dao, db, package_role, user
):
    response = auth_client.delete(
        f"/api/channels/{public_channel.name}/packages/{public_package.name}"
    )

    if package_role == MEMBER:
        assert response.status_code == 403
        return

    package = (
        db.query(Package).filter(Package.name == public_package.name).one_or_none()
    )

    if package_role == MEMBER:
        assert response.status_code == 403
        assert package is not None
    else:
        assert response.status_code == 200
        assert package is None


def test_delete_package_non_member(
    client, public_package, public_channel, dao, db, other_user
):
    response = client.get(f"/api/dummylogin/{other_user.username}")

    assert response.status_code == 200

    response = client.delete(
        f"/api/channels/{public_channel.name}/packages/{public_package.name}"
    )

    assert response.status_code == 403

    package = (
        db.query(Package).filter(Package.name == public_package.name).one_or_none()
    )

    assert package is not None


def test_delete_package_versions_with_package(
    auth_client, public_channel, public_package, package_version, dao, db, pkgstore
):
    assert public_channel.size > 0
    assert public_channel.size == package_version.size

    assert package_version.package_name == public_package.name

    update_indexes(dao, pkgstore, public_channel.name)

    # Get package files
    package_filenames = [
        os.path.join(version.platform, version.filename)
        for version in public_package.package_versions  # type: ignore
    ]

    # Get repodata content
    package_dir = Path(pkgstore.channels_dir) / public_channel.name / 'linux-64'
    with open(package_dir / 'repodata.json', 'r') as fd:
        repodata = json.load(fd)

    # Check that all packages are initially in repodata
    for filename in package_filenames:
        assert os.path.basename(filename) in repodata["packages"].keys()

    # Get channel files
    init_files = sorted(pkgstore.list_files(public_channel.name))

    response = auth_client.delete(
        f"/api/channels/{public_channel.name}/packages/{public_package.name}"
    )

    assert response.status_code == 200

    db.refresh(public_channel)
    assert public_channel.size == 0

    versions = (
        db.query(PackageVersion)
        .filter(PackageVersion.package_name == public_package.name)
        .all()
    )

    assert len(versions) == 0

    # Check that repodata content has been updated
    with open(package_dir / 'repodata.json', 'r') as fd:
        repodata = json.load(fd)

    assert repodata["info"] == repodata["info"]

    # Remove package files from files list
    # Check that packages have been removed from repodata
    for filename in package_filenames:
        init_files.remove(filename)
        assert os.path.basename(filename) not in repodata["packages"]

    # Check that the package tree files is the same except for package files
    files = sorted(pkgstore.list_files(public_channel.name))

    assert files == init_files


def test_get_paginated_package_versions(
    auth_client, public_channel, package_version, dao
):
    response = auth_client.get(
        f"/api/paginated/channels/{public_channel.name}/"
        f"packages/{package_version.package_name}/versions"
    )

    assert response.status_code == 200
    assert isinstance(response.json().get('pagination'), dict)
    assert response.json().get('pagination').get('all_records_count') == 1

    assert isinstance(response.json().get('result'), list)
    assert len(response.json().get('result')) == 1


def test_get_package_version(auth_client, public_channel, package_version, dao):
    filename = "test-package-0.1-0.tar.bz2"
    platform = "linux-64"
    response = auth_client.get(
        f"/api/channels/{public_channel.name}/"
        f"packages/{package_version.package_name}/versions/{platform}/{filename}"
    )

    assert response.status_code == 200
    assert response.json()["filename"] == filename
    assert response.json()["platform"] == platform
    assert response.json()["download_count"] == 0


@pytest.mark.parametrize("user_server_role", [OWNER, MAINTAINER])
@pytest.mark.parametrize("user_package_role", [OWNER, MAINTAINER, MEMBER, None])
@pytest.mark.parametrize("user_channel_role", [OWNER, MAINTAINER, MEMBER, None])
@pytest.mark.parametrize("private", [True, False])
def test_get_package_version_permissions(
    auth_client,
    user,
    private_package_version,
    user_package_role,
    user_channel_role,
    private_channel,
    db,
    private_package,
    private,
    user_server_role,
):
    private_channel.private = private
    user.role = user_server_role

    if user_channel_role:
        channel_member = ChannelMember(
            channel=private_channel, user=user, role=user_channel_role
        )
        db.add(channel_member)
    if user_package_role:
        package_member = PackageMember(
            channel=private_channel,
            user=user,
            package=private_package,
            role=user_package_role,
        )
        db.add(package_member)
    db.commit()

    filename = private_package_version.filename
    platform = private_package_version.platform
    channel_name = private_package_version.channel_name
    package_name = private_package_version.package_name
    response = auth_client.get(
        f"/api/channels/{channel_name}/"
        f"packages/{package_name}/versions/{platform}/{filename}"
    )

    if not private:
        assert response.status_code == 200
    elif user_server_role in [OWNER, MAINTAINER]:
        assert response.status_code == 200
    elif user_channel_role in [OWNER, MAINTAINER, MEMBER]:
        assert response.status_code == 200
    elif user_package_role in [OWNER, MAINTAINER, MEMBER]:
        assert response.status_code == 200
    else:
        assert response.status_code == 403


@pytest.mark.parametrize("user_server_role", [OWNER, MAINTAINER])
@pytest.mark.parametrize("user_package_role", [OWNER, MAINTAINER, MEMBER, None])
@pytest.mark.parametrize("user_channel_role", [OWNER, MAINTAINER, MEMBER, None])
@pytest.mark.parametrize("private", [True, False])
def test_delete_package_version_permissions(
    auth_client,
    user,
    private_package_version,
    user_package_role,
    user_channel_role,
    private_channel,
    db,
    private_package,
    pkgstore,
    private,
    user_server_role,
):
    private_channel.private = private
    user.role = user_server_role

    if user_channel_role:
        channel_member = ChannelMember(
            channel=private_channel, user=user, role=user_channel_role
        )
        db.add(channel_member)
    if user_package_role:
        package_member = PackageMember(
            channel=private_channel,
            user=user,
            package=private_package,
            role=user_package_role,
        )
        db.add(package_member)
    db.commit()

    filename = private_package_version.filename
    platform = private_package_version.platform
    channel_name = private_package_version.channel_name
    package_name = private_package_version.package_name
    response = auth_client.delete(
        f"/api/channels/{channel_name}/"
        f"packages/{package_name}/versions/{platform}/{filename}"
    )

    if user_server_role in [OWNER, MAINTAINER]:
        assert response.status_code == 200
    elif user_channel_role in [OWNER, MAINTAINER]:
        assert response.status_code == 200
    elif user_package_role in [OWNER, MAINTAINER]:
        assert response.status_code == 200
    else:
        assert response.status_code == 403


def test_get_non_existing_package_version(
    auth_client, public_channel, package_version, dao
):
    filename = "test-package-0.2-0.tar.bz2"
    platform = "linux-64"
    response = auth_client.get(
        f"/api/channels/{public_channel.name}/"
        f"packages/test-package/versions/{platform}/{filename}"
    )

    assert response.status_code == 404


@pytest.fixture
def remove_package_versions(db):
    yield
    db.query(PackageVersion).delete()


@pytest.mark.parametrize("package_name", ["test-package", "my-package"])
def test_upload_package_version(
    auth_client,
    public_channel,
    public_package,
    package_name,
    db,
    config,
    remove_package_versions,
):
    pkgstore = config.get_package_store()

    package_filename = "test-package-0.1-0.tar.bz2"
    with open(package_filename, "rb") as fid:
        files = {"files": (package_filename, fid)}
        response = auth_client.post(
            f"/api/channels/{public_channel.name}/packages/"
            f"{public_package.name}/files/",
            files=files,
        )

    with open(package_filename, "rb") as fid:
        condainfo = CondaInfo(fid, package_filename)
        condainfo._parse_conda()

    package_dir = Path(pkgstore.channels_dir) / public_channel.name / 'linux-64'

    if package_name == "my-package":
        assert response.status_code == 400
        detail = response.json()['detail']
        assert "package endpoint" in detail
        assert "does not match" in detail
        assert "test-package" in detail
        assert "my-package" in detail
        assert package_filename not in os.listdir(package_dir)
    else:
        assert response.status_code == 201
        db.refresh(public_channel)
        assert public_channel.size == condainfo.info['size']
        assert pkgstore.serve_path(
            public_channel.name, str(Path(condainfo.info['subdir']) / package_filename)
        )
        assert package_filename in os.listdir(package_dir)


def _upload_file_1(
    auth_client,
    public_channel,
    public_package,
    filepath: Path,
    force: bool = False,
):
    """Upload a file using /channels/{channel_name}/packages/{package_name}/files/"""
    with open(filepath, "rb") as fid:
        files = {"files": (filepath.name, fid)}
        response = auth_client.post(
            f"/api/channels/{public_channel.name}/packages/"
            f"{public_package.name}/files/",
            files=files,
            data={"force": force},
        )
    return response


def _upload_file_2(
    auth_client,
    public_channel,
    public_package,
    filepath: Path,
    force: bool = False,
):
    """Upload a file using /channels/{channel_name}/upload/{file_name}"""

    with open(filepath, "rb") as fid:
        body_bytes = fid.read()
    response = auth_client.post(
        f"/api/channels/{public_channel.name}/upload/{filepath.name}",
        content=body_bytes,
        params={"force": force, "sha256": hashlib.sha256(body_bytes).hexdigest()},
    )
    return response


@pytest.mark.parametrize("upload_function", [_upload_file_1, _upload_file_2])
def test_upload_package_version_wrong_filename(
    auth_client,
    public_channel,
    public_package,
    package_name,
    db,
    config,
    remove_package_versions,
    upload_function: Callable,
):
    pkgstore = config.get_package_store()

    package_filename = "my-package-0.1-0.tar.bz2"
    os.rename("test-package-0.1-0.tar.bz2", package_filename)

    response = upload_function(
        auth_client, public_channel, public_package, Path(package_filename)
    )

    package_dir = Path(pkgstore.channels_dir) / public_channel.name / 'linux-64'

    assert response.status_code == 400
    detail = response.json()['detail']
    assert "info file" in detail
    assert "do not match" in detail
    assert "my-package" in detail
    assert not os.path.exists(package_dir)


def sha_and_md5(path: Union[Path, str]) -> Tuple[str, str]:
    sha = hashlib.sha256()
    md5 = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(2**16), b""):
            sha.update(chunk)
            md5.update(chunk)
    return sha.hexdigest(), md5.hexdigest()


@pytest.mark.parametrize("upload_function", [_upload_file_1, _upload_file_2])
@pytest.mark.parametrize("package_name", ["test-package"])
def test_upload_duplicate_package_version(
    auth_client,
    public_channel,
    public_package,
    package_name,
    db,
    config,
    remove_package_versions,
    upload_function: Callable,
    tmp_path,
):
    pkgstore = config.get_package_store()

    def get_repodata():
        """Helper function to read repo data"""
        package_dir = Path(pkgstore.channels_dir) / public_channel.name / 'linux-64'
        return json.loads((package_dir / 'repodata.json').read_text())

    # Test setup: path1 is a package we will upload
    path1 = Path(__file__).parent.parent / "data" / "test-package-0.1-0.tar.bz2"

    # To test duplicate uploads, we have a second file `_copy`, which is the same
    # package and version but has a different content, so different hashes
    # We must move this file to a temporary directory so that we can give it the same
    # name as the first file.
    path2 = tmp_path / path1.name
    shutil.copyfile(
        Path(__file__).parent.parent / "data" / "test-package-0.1-0_copy.tar.bz2", path2
    )

    # Sanity checks
    sha1, md51 = sha_and_md5(path1)
    sha2, md52 = sha_and_md5(path2)
    assert (sha1 != sha2) and (
        md51 != md52
    ), "Sanity check failure: Test files have same hash"
    assert (
        path1.name == path2.name
    ), "Sanity check failure: Test files have different name"

    size1 = os.path.getsize(path1)
    size2 = os.path.getsize(path2)
    assert (
        size1 != size2
    ), "Sanity check failure: Test files must have different sizes for this test."

    # First upload
    # File should not exist
    assert not pkgstore.file_exists(public_channel.name, f"linux-64/{path1.name}")
    response = upload_function(auth_client, public_channel, public_package, path1)

    # Expect success
    assert response.status_code == 201

    # Check meta data is OK
    repodata_after_first = get_repodata()
    assert repodata_after_first["packages"][path1.name]["sha256"] == sha1
    assert repodata_after_first["packages"][path1.name]["md5"] == md51
    assert repodata_after_first["packages"][path1.name]["size"] == size1

    # Check that the file in the store is OK
    file_in_store = (
        Path(pkgstore.channels_dir) / public_channel.name / 'linux-64' / path1.name
    )
    assert size1 == os.path.getsize(file_in_store)
    assert (sha1, md51) == sha_and_md5(file_in_store)

    # Second upload: File with same name but different content, without force
    response = upload_function(auth_client, public_channel, public_package, path2)

    # Expect 409 since the file already exists
    assert response.status_code == 409
    detail = response.json()['detail']
    assert "Duplicate" in detail

    # Check meta data is OK: It should not have changed with respect to before
    repodata_after_second = get_repodata()
    assert repodata_after_second == repodata_after_first

    # Check that the file in the store is OK
    file_in_store = (
        Path(pkgstore.channels_dir) / public_channel.name / 'linux-64' / path1.name
    )

    # File in store should  not be the second file
    assert not (
        (sha2, md52) == sha_and_md5(file_in_store)
    ), "Duplicate upload without force updated stored file."
    assert not (
        size2 == os.path.getsize(file_in_store)
    ), "Duplicate upload without force updated stored file."

    # File in store should be the first file
    assert size1 == os.path.getsize(
        file_in_store
    ), "Duplicate upload without force updated stored file."
    assert (sha1, md51) == sha_and_md5(
        file_in_store
    ), "Duplicate upload without force updated stored file."

    # Third upload: Same as second but now with force
    # Ensure the 'time_modified' value changes in repodata.json
    time.sleep(1)

    # Submit the same package with 'force' flag
    response = upload_function(
        auth_client, public_channel, public_package, path2, force=True
    )
    assert response.status_code == 201

    # Check that repodata content has been updated
    repodata_after_force = get_repodata()

    # Info should match
    assert repodata_after_first["info"] == repodata_after_force["info"]

    # Package keys should match
    assert (
        repodata_after_first["packages"].keys()
        == repodata_after_force["packages"].keys()
    )

    # Hashes should now have changed to the second file
    assert repodata_after_force["packages"][path1.name]["sha256"] == sha2
    assert repodata_after_force["packages"][path1.name]["md5"] == md52
    assert repodata_after_force["packages"][path1.name]["size"] == size2

    assert (
        repodata_after_force["packages"][path1.name]["time_modified"]
        > repodata_after_first["packages"][path1.name]["time_modified"]
    )


@pytest.mark.parametrize("package_name", ["test-package"])
def test_check_channel_size_limits(
    auth_client, public_channel, public_package, db, config
):
    public_channel.size_limit = 0
    db.commit()
    pkgstore = config.get_package_store()

    package_filename = "test-package-0.1-0.tar.bz2"
    with open(package_filename, "rb") as fid:
        files = {"files": (package_filename, fid)}
        response = auth_client.post(
            f"/api/channels/{public_channel.name}/packages/"
            f"{public_package.name}/files/",
            files=files,
        )

    assert response.status_code == 422
    detail = response.json()['detail']
    assert "quota" in detail
    with pytest.raises(FileNotFoundError):
        pkgstore.serve_path(
            public_channel.name, str(Path("linux-64") / package_filename)
        )


def test_delete_package_version(
    auth_client, public_channel, package_version, dao, pkgstore, db
):
    assert public_channel.size > 0
    assert public_channel.size == package_version.size

    filename = "test-package-0.1-0.tar.bz2"
    platform = "linux-64"

    update_indexes(dao, pkgstore, public_channel.name)

    # Get repodata content and check that package is inside
    package_dir = Path(pkgstore.channels_dir) / public_channel.name / 'linux-64'
    with open(package_dir / 'repodata.json', 'r') as fd:
        repodata = json.load(fd)
    assert filename in repodata["packages"].keys()

    response = auth_client.delete(
        f"/api/channels/{public_channel.name}/"
        f"packages/{package_version.package_name}/versions/{platform}/{filename}"
    )

    assert response.status_code == 200

    versions = (
        db.query(PackageVersion)
        .filter(PackageVersion.package_name == package_version.package_name)
        .all()
    )

    assert len(versions) == 0

    with pytest.raises(Exception):
        pkgstore.serve_path(public_channel.name, str(Path(platform) / filename))

    db.refresh(public_channel)
    assert public_channel.size == 0

    # Check that repodata content has been updated
    with open(package_dir / 'repodata.json', 'r') as fd:
        repodata = json.load(fd)
    assert filename not in repodata["packages"].keys()


def test_package_name_length_limit(auth_client, public_channel, db):
    package_name = "package_" * 100

    response = auth_client.post(
        f"/api/channels/{public_channel.name}/packages", json={"name": package_name}
    )

    assert response.status_code == 201

    pkg = db.query(Package).filter(Package.name == package_name).one_or_none()

    assert pkg is not None


def test_validate_package_names(auth_client, public_channel, remove_package_versions):
    valid_package_names = [
        "interesting-package",
        "valid.package.name",
        "valid-package-name",
        "valid_package_name",
        "validpackage1234",
    ]

    for package_name in valid_package_names:
        response = auth_client.post(
            f"/api/channels/{public_channel.name}/packages", json={"name": package_name}
        )

        assert response.status_code == 201

    invalid_package_names = [
        "InvalidPackage",  # no uppercase
        "invalid%20package",  # no spaces
        "invalid package",  # no spaces
        "invalid%package",  # no special characters
        "**invalidpackage**",
        "błędnypakiet",  # no unicode
    ]

    for package_name in invalid_package_names:
        response = auth_client.post(
            f"/api/channels/{public_channel.name}/packages", json={"name": package_name}
        )
        assert response.status_code == 422


@pytest.mark.parametrize(
    "package_name,msg",
    [
        ("TestPackage", "String should match"),
        ("test-package", None),
    ],
)
def test_validate_package_names_files_endpoint(
    auth_client,
    public_channel,
    mocker,
    package_name,
    msg,
    config: Config,
    remove_package_versions,
):
    pkgstore = config.get_package_store()

    package_filename = "test-package-0.1-0.tar.bz2"

    with open(package_filename, "rb") as fid:
        condainfo = CondaInfo(fid, package_filename)
        condainfo._parse_conda()

    # patch conda info
    condainfo.info['name'] = package_name
    condainfo.channeldata['packagename'] = package_name

    mocked_cls = mocker.patch("quetz.main.CondaInfo")
    mocked_cls.return_value = condainfo

    with open(package_filename, "rb") as fid:
        files = {"files": (f"{package_name}-0.1-0.tar.bz2", fid)}
        response = auth_client.post(
            f"/api/channels/{public_channel.name}/files/", files=files
        )

    if msg:
        assert response.status_code == 422
        assert msg in response.json()["detail"]

        with pytest.raises(FileNotFoundError):
            pkgstore.serve_path(
                public_channel.name, f'linux-64/{package_name}-0.1-0.tar.bz2'
            )
    else:
        assert response.status_code == 201
        assert pkgstore.serve_path(
            public_channel.name, f'linux-64/{package_name}-0.1-0.tar.bz2'
        )


@pytest.fixture
def plugin(app):
    from quetz.main import pm

    class Plugin:
        @hookimpl
        def validate_new_package(
            self,
            channel_name: str,
            package_name: str,
            file_handler: BinaryIO,
            condainfo: CondaInfo,
        ):
            raise ValidationError(f"name {package_name} not allowed")

    plugin = Plugin()
    pm.register(plugin)
    yield plugin
    pm.unregister(plugin)


def test_validation_hook(auth_client, public_channel, plugin, config):
    pkgstore = config.get_package_store()

    response = auth_client.post(
        f"/api/channels/{public_channel.name}/packages", json={"name": "package-name"}
    )

    assert response.status_code == 422
    assert "package-name not allowed" in response.json()["detail"]

    package_filename = "test-package-0.1-0.tar.bz2"
    with open(package_filename, "rb") as fid:
        files = {"files": (package_filename, fid)}
        response = auth_client.post(
            f"/api/channels/{public_channel.name}/files/", files=files
        )

    assert response.status_code == 422
    assert "test-package not allowed" in response.json()["detail"]
    with pytest.raises(FileNotFoundError):
        pkgstore.serve_path(public_channel.name, 'linux-64/test-package-0.1-0.tar.bz2')


@pytest.mark.parametrize(
    "endpoint",
    [
        "/api/channels/{channel_name}/packages/{package_name}",
        "/api/channels/{channel_name}/packages",
        "/api/packages/search/?q='channel:{channel_name}'",
    ],
)
@pytest.mark.parametrize("package_name", ["test-package"])
def test_package_current_version(
    auth_client, make_package_version, channel_name, endpoint
):
    # test platforms, current_version and url
    make_package_version("test-package-0.1-0.tar.bz2", "0.1", platform="linux-64")
    v1n = make_package_version("test-package-0.1-0.tar.bz2", "0.1", platform="noarch")
    v2l = make_package_version("test-package-0.2-0.tar.bz2", "0.2", platform="linux-64")
    v2o = make_package_version("test-package-0.2-0.tar.bz2", "0.2", platform="os-x")

    response = auth_client.get(
        endpoint.format(channel_name=channel_name, package_name=v2o.package_name)
    )
    assert response.status_code == 200

    package_data = response.json()
    if isinstance(package_data, list):
        assert len(package_data) == 1
        package_data = package_data[0]
    assert package_data['current_version'] == v2o.version

    # delete v0.2 os-x and linux-64 package
    response = auth_client.delete(
        f"/api/channels/{channel_name}/"
        f"packages/{v2o.package_name}/versions/{v2o.platform}/{v2o.filename}"
    )
    assert response.status_code == 200
    response = auth_client.delete(
        f"/api/channels/{channel_name}/"
        f"packages/{v2l.package_name}/versions/{v2l.platform}/{v2l.filename}"
    )
    assert response.status_code == 200

    # require current version to be 0.1
    response = auth_client.get(
        endpoint.format(channel_name=channel_name, package_name=v1n.package_name)
    )
    assert response.status_code == 200

    package_data = response.json()
    if isinstance(package_data, list):
        assert len(package_data) == 1
        package_data = package_data[0]
    assert package_data['current_version'] == v1n.version


@pytest.mark.parametrize("package_name", ["test-package"])
def test_get_package_with_versions(
    make_package_version, channel_name, dao, package_name
):
    # test loading of latest (current) and all other versions

    make_package_version("test-package-0.1-0.tar.bz2", "0.1", platform="linux-64")
    make_package_version("test-package-0.1-0.tar.bz2", "0.1", platform="noarch")
    v = make_package_version("test-package-0.2-0.tar.bz2", "0.2", platform="linux-64")

    package = dao.get_package(channel_name, package_name)

    assert package.current_package_version == v
    assert len(package.package_versions) == 3


@pytest.mark.parametrize("package_name", ["test-package"])
def test_package_channel_data_attributes(
    auth_client,
    make_package_version,
    channel_name,
    remove_package_versions,
):
    # test attributes derived from channel data
    for package_filename in Path(".").glob("*.tar.bz2"):
        with open(package_filename, "rb") as fid:
            files = {"files": (str(package_filename), fid)}
            response = auth_client.post(
                f"/api/channels/{channel_name}/files/",
                files=files,
            )

    response = auth_client.get(f"/api/channels/{channel_name}/packages/test-package")
    assert response.status_code == 200
    content = response.json()
    assert content['platforms'] == ['linux-64']
    assert content['url'].startswith("https://")


@pytest.fixture
def owner(db, dao: Dao):
    # create an owner with OWNER role
    owner = dao.create_user_with_role("owner", role=SERVER_OWNER)

    yield owner

    db.delete(owner)
    db.commit()


@pytest.fixture
def private_channel(db, dao: Dao, owner):
    # create a channel
    channel_data = Channel(name='private-channel', private=True)
    channel = dao.create_channel(channel_data, owner.id, "owner")

    yield channel

    db.delete(channel)
    db.commit()


@pytest.fixture
def private_package(db, owner, private_channel, dao):
    package_data = Package(name='test-package')

    package = dao.create_package(private_channel.name, package_data, owner.id, "owner")

    yield package

    db.delete(package)
    db.commit()


@pytest.fixture
def api_key(db, dao: Dao, owner, private_channel):
    # create an api key with restriction
    key = dao.create_api_key(
        owner.id,
        BaseApiKey.model_validate(
            dict(
                description="test api key",
                expire_at="2099-12-31",
                roles=[
                    {
                        'role': 'maintainer',
                        'package': None,
                        'channel': private_channel.name,
                    }
                ],
            )
        ),
        "API key with role restruction",
    )

    yield key

    # delete API Key
    key.deleted = True
    db.commit()


def test_upload_package_with_api_key(client, dao: Dao, owner, private_channel, api_key):
    # set api key of the anonymous user 'owner_key' to headers
    client.headers['X-API-Key'] = api_key.key

    # post new package
    response = client.post(
        f"/api/channels/{private_channel.name}/packages",
        json={"name": "test-package", "summary": "none", "description": "none"},
    )
    assert response.status_code == 201

    # we used the anonymous user of the API key for upload,
    # but expect the owner to be the package owner
    member = dao.get_package_member(
        private_channel.name, 'test-package', username=owner.username
    )
    assert member


def test_upload_file_to_package_with_api_key(
    dao: Dao, client, owner, private_channel, private_package, api_key
):
    # set api key of the anonymous user 'owner_key' to headers
    client.headers['X-API-Key'] = api_key.key

    package_filename = "test-package-0.1-0.tar.bz2"
    with open(package_filename, "rb") as fid:
        files = {"files": (package_filename, fid)}
        response = client.post(
            f"/api/channels/{private_channel.name}/packages/"
            f"{private_package.name}/files/",
            files=files,
        )
        assert response.status_code == 201

    # we used the anonymous user of the API key for upload,
    # but expect the owner to be the package owner
    version = dao.get_package_version_by_filename(
        channel_name=private_channel.name,
        package_name=private_package.name,
        filename=package_filename,
        platform='linux-64',
    )
    assert version
    assert version.uploader == owner


def test_upload_file_to_channel_with_api_key(
    dao: Dao, client, owner, private_channel, api_key
):
    # set api key of the anonymous user 'owner_key' to headers
    client.headers['X-API-Key'] = api_key.key

    package_filename = "test-package-0.1-0.tar.bz2"
    with open(package_filename, "rb") as fid:
        files = {"files": (package_filename, fid)}
        response = client.post(
            f"/api/channels/{private_channel.name}/files/",
            files=files,
        )
        assert response.status_code == 201

    # we used the anonymous user of the API key for upload,
    # but expect the owner to be the package owner
    version = dao.get_package_version_by_filename(
        channel_name=private_channel.name,
        package_name='test-package',
        filename=package_filename,
        platform='linux-64',
    )
    assert version
    assert version.uploader == owner



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

import json
import logging
import uuid
from collections import defaultdict
from datetime import date, datetime
from itertools import groupby
from typing import TYPE_CHECKING, Dict, List, Optional

from sqlalchemy import and_, func, insert, or_
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.exc import IntegrityError, NoResultFound  # type: ignore
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import Query, Session, aliased, joinedload
from sqlalchemy.sql.expression import FunctionElement, Insert
from sqlalchemy.types import DateTime

from quetz import channel_data, errors, rest_models, versionorder
from quetz.database_extensions import version_match
from quetz.utils import apply_custom_query

from .db_models import (
    ApiKey,
    Channel,
    ChannelMember,
    ChannelMirror,
    Email,
    Identity,
    Package,
    PackageMember,
    PackageVersion,
    Profile,
    User,
)
from .jobs.models import Job, JobStatus, Task, TaskStatus
from .metrics.db_models import (
    IntervalType,
    PackageVersionMetric,
    next_timestamp,
    round_timestamp,
)

if TYPE_CHECKING:
    from quetz.authentication import base as auth_base

logger = logging.getLogger("quetz")


class date_trunc(FunctionElement):
    """round timestamp to nearest starting edge of an interval

    Arguments
    ---------

    interval: IntervalEnum

    timestamp: datetime
    """

    name = "date_trunc"
    type = DateTime()


@compiles(date_trunc, 'postgresql')
def pg_date_trunc(element, compiler, **kw):
    pg_map = {"H": "hour", "D": "day", "M": "month", "Y": "year"}
    period, date = list(element.clauses)
    return f"date_trunc('{pg_map[period.value.value]}', {compiler.process(date, **kw)})"


@compiles(date_trunc, 'sqlite')
def sqlite_date_trunc(element, compiler, **kw):
    period, date = list(element.clauses)
    now_interval = round_timestamp(date.value, period.value)
    date.value = now_interval
    return compiler.process(date)


class Upsert(Insert):
    """Upsert for PackageVersionMetrics table. Requires a unique
    constraint to be defined.

    Arguments
    ---------

    table

    values: values to insert

    index_elements: columns of unique constraint

    column: Column to be incremented

    incr: increment
    """

    inherit_cache = False

    def __init__(self, table, values, index_elements, column, incr=1):
        self.values = values
        self.index_elements = index_elements
        self.column = column
        self._returning = None
        self.table = table
        self.incr = incr
        self._inline = False


@compiles(Upsert, 'postgresql')
def upsert_pg(element, compiler, **kw):
    index_elements = element.index_elements
    values = element.values
    column = element.column
    incr = element.incr
    table = element.table

    stmt = pg_insert(table).values(values)
    stmt = stmt.on_conflict_do_update(
        index_elements=index_elements,
        set_={column.name: column + incr},
    )

    return compiler.visit_insert(stmt)


@compiles(Upsert, 'sqlite')
def upsert_sql(element, compiler, **kw):
    # on_conflict_do_update does exist in sqlite
    # but it was ported to sqlalchemy only in version 1.4
    # which was not released at the time of implementing this
    # so we treat it with raw SQL syntax
    # sqlite ref: https://www.sqlite.org/lang_upsert.html
    # sqlalchemy 1.4 ref: https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#insert-on-conflict-upsert # noqa

    index_elements = element.index_elements
    values = element.values
    column = element.column
    incr = element.incr
    table = element.table

    stmt = insert(table).values(values)
    raw_sql = compiler.process(stmt)
    upsert_stmt = (
        f"ON CONFLICT ({','.join(index_elements)}) "
        f"DO UPDATE SET {column.name}={column.name}+{incr}"
    )

    return raw_sql + " " + upsert_stmt


def get_paginated_result(query: Query, skip: int, limit: int):
    count = query.order_by(None).count()
    query = query.offset(skip)
    if limit >= 0:
        query = query.limit(limit)
    return {
        'pagination': {'skip': skip, 'limit': limit, 'all_records_count': count},
        'result': query.all(),
    }


def _parse_sort_by(query, model, sortstr: str):
    sorts = sortstr.split(',')

    for s in sorts:
        splitted = s.split(':')
        if len(splitted) == 2:
            field, order = splitted
        else:
            field = s
            order = 'desc'

        if hasattr(model, field):
            model_field = getattr(model, field)
            if order == 'desc':
                query = query.order_by(model_field.desc())
            else:
                query = query.order_by(model_field.asc())
    return query


class Dao:
    db: Session

    def __init__(self, db: Session):
        self.db = db

    def rollback(self):
        self.db.rollback()

    def get_profile(self, user_id):
        try:
            return self.db.query(Profile).filter(Profile.user_id == user_id).one()
        except NoResultFound:
            logger.error("User not found")

    def get_user(self, user_id):
        try:
            return self.db.query(User).filter(User.id == user_id).one()
        except NoResultFound:
            logger.error("User not found")

    def get_users(self, skip: int, limit: int, q: str, order_by: str = 'username:asc'):
        query = (
            self.db.query(User)
            .filter(User.username.isnot(None))
            .filter(User.profile.has())
        )

        if q:
            query = query.filter(User.username.ilike(f'%{q}%'))

        query = query.options(joinedload(User.profile))

        if order_by:
            query = _parse_sort_by(query, User, order_by)

        if limit < 0:
            return query.all()

        return get_paginated_result(query, skip, limit)

    def get_user_by_username(self, username: str) -> Optional[User]:
        return (
            self.db.query(User)
            .filter(User.username == username)
            .options(joinedload(User.profile))
            .one_or_none()
        )

    def delete_user(self, user_id: bytes):
        # we are not really removing users
        # only their identity providers and profiles
        self.db.query(Profile).filter(Profile.user_id == user_id).delete()
        self.db.query(Identity).filter(Identity.user_id == user_id).delete()
        self.db.query(ApiKey).filter(
            or_(ApiKey.user_id == user_id, ApiKey.owner_id == user_id)
        ).delete()
        self.db.commit()

    def set_user_role(self, username: str, role: Optional[str]):
        user = self.db.query(User).filter(User.username == username).one_or_none()

        if user:
            user.role = role
            self.db.commit()

    def get_channels(
        self,
        skip: int,
        limit: int,
        q: Optional[str],
        user_id: Optional[bytes],
        include_public: bool = True,
    ):
        query = self.db.query(Channel)

        if user_id:
            if include_public:
                query = query.filter(
                    or_(
                        ~Channel.private,
                        Channel.members.any(ChannelMember.user_id == user_id),
                    )
                )
            else:
                query = query.filter(
                    Channel.members.any(ChannelMember.user_id == user_id)
                )
        else:
            query = query.filter(Channel.private == False)  # noqa

        if q:
            query = query.filter(Channel.name.ilike(f'%{q}%'))

        if limit < 0:
            return query.all()

        return get_paginated_result(query, skip, limit)

    def get_user_channels_with_role(
        self,
        skip: int,
        limit: int,
        user_id: bytes,
    ):
        query = (
            self.db.query(Channel.name, ChannelMember.role)
            .join(ChannelMember)
            .filter(ChannelMember.user_id == user_id)
        )

        if limit < 0:
            return query.all()

        return get_paginated_result(query, skip, limit)

    def create_channel(
        self,
        data: rest_models.Channel,
        user_id: Optional[bytes] = None,
        role: Optional[str] = None,
        size_limit: Optional[int] = None,
    ):
        if '_' in data.name:
            raise errors.ValidationError("_ should not be used in channel name")
        if not data.name.isascii():
            raise errors.ValidationError(
                "only ASCII characters should be used in channel name"
            )

        channel = Channel(
            name=data.name,
            description=data.description,
            mirror_channel_url=data.mirror_channel_url,
            mirror_mode=data.mirror_mode,
            private=data.private,
            ttl=data.ttl,
            channel_metadata=json.dumps(data.metadata.__dict__),
            size_limit=size_limit,
        )

        self.db.add(channel)

        if role and user_id:
            member = ChannelMember(channel=channel, user_id=user_id, role=role)
            self.db.add(member)

        self.db.commit()

        return channel

    def cleanup_channel_db(
        self,
        channel_name: str,
        package_name: Optional[str] = None,
        dry_run: bool = False,
    ):
        # remove all Packages without PackageVersions
        package_without_package_versions = []
        all_packages = self.db.query(Package).filter(
            Package.channel_name == channel_name
        )
        if package_name:
            all_packages = all_packages.filter(
                PackageVersion.package_name == package_name
            )
        for each_package in all_packages:
            all_package_versions = (
                self.db.query(PackageVersion)
                .filter(PackageVersion.channel_name == channel_name)
                .filter(PackageVersion.package_name == each_package.name)
            )
            if all_package_versions.count() == 0:
                package_without_package_versions.append(each_package.name)

        for each_package_name in package_without_package_versions:
            if not dry_run:
                self.db.query(PackageMember).filter(
                    PackageMember.channel_name == channel_name
                ).filter(PackageMember.package_name == each_package_name).delete()

                self.db.query(Package).filter(
                    Package.channel_name == channel_name
                ).filter(Package.name == each_package_name).delete()

            logger.info(
                f"removing Package {channel_name}/{each_package_name} from db as "
                "it has no PackageVersions"
            )

        if not dry_run:
            self.db.commit()

        # clean platforms / channeldata for Packages
        all_packages = self.db.query(Package).filter(
            Package.channel_name == channel_name
        )
        if package_name:
            all_packages = all_packages.filter(
                PackageVersion.package_name == package_name
            )
        for each_package in all_packages:
            if each_package.channeldata is not None:
                each_package_channeldata = json.loads(each_package.channeldata)
                subdirs = each_package_channeldata["subdirs"]
                if subdirs:
                    if not dry_run:
                        subdirs = sorted(list(set(subdirs)))
                        each_package_channeldata["subdirs"] = subdirs
                        each_package.channeldata = json.dumps(each_package_channeldata)
                        each_package.url = each_package_channeldata.get("home", "")
                        each_package.platforms = ":".join(
                            each_package_channeldata.get("subdirs", [])
                        )
                    logger.info(
                        "cleaning platforms and "
                        f"channeldata for {channel_name}/{each_package.name}"
                    )

        if not dry_run:
            self.db.commit()
        if not package_name:
            logger.info(f"Done cleaning up db for {channel_name}")
        else:
            logger.info(f"Done cleaning up db for {channel_name}/{package_name}")

        # Re-sort all PackageVersions
        all_packages = self.db.query(Package).filter(
            Package.channel_name == channel_name
        )
        if package_name:
            all_packages = all_packages.filter(
                PackageVersion.package_name == package_name
            )
        for x, each_package in enumerate(all_packages):
            all_versions_for_each_package = (
                self.db.query(PackageVersion)
                .filter(PackageVersion.channel_name == channel_name)
                .filter(PackageVersion.package_name == each_package.name)
                .order_by(PackageVersion.version_order.asc())
            ).all()

            if not dry_run:
                v_dict = {}
                for each_package_version in all_versions_for_each_package:
                    if each_package_version.version is not None:
                        v_dict[each_package_version] = versionorder.VersionOrder(
                            each_package_version.version
                        )
                sorted_v = sorted(
                    v_dict.items(), key=lambda item: item[1], reverse=True
                )

                for i, (each_package_version, version) in enumerate(sorted_v):
                    each_package_version.version_order = i
            logger.info(
                f"Re-sorted PackageVersions for {channel_name}/{each_package.name}"
            )

        if not dry_run:
            self.db.commit()
        if not package_name:
            logger.info(f"Done sorting package versions for {channel_name}")
        else:
            logger.info(
                f"Done sorting package versions for {channel_name}/{package_name}"
            )

    def create_channel_mirror(
        self, channel_name: str, url: str, api_endpoint: str, metrics_endpoint: str
    ):
        channel_mirror = ChannelMirror(
            channel_name=channel_name,
            url=url,
            api_endpoint=api_endpoint,
            metrics_endpoint=metrics_endpoint,
        )
        self.db.add(channel_mirror)
        self.db.commit()

        return channel_mirror

    def delete_channel_mirror(self, channel_name: str, mirror_id: str):
        mirror_uuid = uuid.UUID(mirror_id).bytes
        self.db.query(ChannelMirror).filter(ChannelMirror.id == mirror_uuid).filter(
            ChannelMirror.channel_name == channel_name
        ).delete()
        self.db.commit()

    def update_channel(self, channel_name, data: dict):
        self.db.query(Channel).filter(Channel.name == channel_name).update(
            data, synchronize_session=False
        )
        self.db.commit()

    def delete_channel(self, channel_name):
        channel = self.get_channel(channel_name)

        self.db.delete(channel)
        self.db.commit()

    def get_packages(
        self,
        channel_name: str,
        skip: int,
        limit: int,
        q: Optional[str] = None,
        order_by: Optional[str] = None,
    ):
        query = self.db.query(Package).filter(Package.channel_name == channel_name)

        if q:
            query = query.filter(Package.name.like(f'%{q}%'))

        if limit < 0:
            return query.all()

        if not order_by:
            query = query.order_by(Package.name.asc())
        else:
            orders = order_by.split(',')
            for o in orders:
                if o.startswith('latest_change'):
                    query = query.join(Package.current_package_version)
                    if len(o.split(':')) == 2 and o.split(':') == 'desc':
                        query = query.order_by(PackageVersion.time_created.desc())
                    else:
                        query = query.order_by(PackageVersion.time_created.asc())
                else:
                    query = _parse_sort_by(query, Package, order_by)

        return get_paginated_result(query, skip, limit)

    def get_user_packages(self, skip: int, limit: int, user_id: bytes):
        query = (
            self.db.query(Package.name, Package.channel_name, PackageMember.role)
            .join(PackageMember)
            .filter(PackageMember.user_id == user_id)
        )

        if limit < 0:
            return query.all()

        return get_paginated_result(query, skip, limit)

    def search_packages(
        self,
        keywords: List[str],
        filters: Optional[List[tuple]],
        user_id: Optional[bytes],
        order_by: str = 'name:asc',
    ):
        db = self.db.query(Package).join(Channel)
        query = apply_custom_query('package', db, keywords, filters)
        if user_id:
            query = query.filter(
                or_(
                    Channel.private == False,  # noqa
                    Channel.members.any(ChannelMember.user_id == user_id),
                )
            )
        else:
            query = query.filter(Channel.private == False)  # noqa

        if order_by:
            query = _parse_sort_by(query, Package, order_by)

        return query.all()

    def search_channels(
        self,
        keywords: List[str],
        filters: Optional[List[tuple]],
        user_id: Optional[bytes],
        order_by: str = 'name:asc',
    ):
        db = self.db.query(Channel)
        query = apply_custom_query('channel', db, keywords, filters)
        if user_id:
            query = query.filter(
                or_(
                    Channel.private == False,  # noqa
                    Channel.members.any(ChannelMember.user_id == user_id),
                )
            )
        else:
            query = query.filter(Channel.private == False)  # noqa

        if order_by:
            query = _parse_sort_by(query, Channel, order_by)

        return query.all()

    def get_channel(self, channel_name: str) -> Optional[Channel]:
        return self.db.query(Channel).filter(Channel.name == channel_name).one_or_none()

    def get_package(self, channel_name: str, package_name: str) -> Optional[Package]:
        return (
            self.db.query(Package)
            .join(Channel)
            .filter(Channel.name == channel_name)
            .filter(Package.name == package_name)
            .one_or_none()
        )

    def create_package(
        self,
        channel_name: str,
        new_package: rest_models.Package,
        user_id: bytes,
        role: str,
    ):
        package = Package(
            name=new_package.name,
            summary=new_package.summary,
            description=new_package.description,
            channeldata="{}",
        )

        package.channel = (
            self.db.query(Channel).filter(Channel.name == channel_name).one()
        )

        member = PackageMember(
            channel=package.channel, package=package, user_id=user_id, role=role
        )

        self.db.add(package)
        self.db.add(member)

        try:
            self.db.commit()
        except IntegrityError as exc:
            self.db.rollback()
            raise errors.DBError(str(exc))

        return package

    def update_package_channeldata(self, channel_name, package_name, channeldata):
        package = self.get_package(channel_name, package_name)
        if package is None:
            raise KeyError(f"Package '{package_name}' not found.")
        if package.channeldata:
            old_data = json.loads(package.channeldata)
        else:
            old_data = None
        data = channel_data.combine(old_data, channeldata)
        package.channeldata = json.dumps(data)
        package.url = data.get("home", "")
        package.platforms = ":".join(data.get("subdirs", []))
        self.db.commit()

    def get_channel_members(self, channel_name: str):
        return (
            self.db.query(ChannelMember)
            .join(User)
            .filter(User.username.isnot(None))
            .filter(ChannelMember.channel_name == channel_name)
            .all()
        )

    def get_channel_member(self, channel_name, username):
        return (
            self.db.query(ChannelMember)
            .join(User)
            .filter(ChannelMember.channel_name == channel_name)
            .filter(User.username == username)
            .one_or_none()
        )

    def create_channel_member(self, channel_name, new_member):
        user = self.get_user_by_username(new_member.username)
        if user is None:
            raise KeyError(f"User '{new_member.username}' not found.")

        member = ChannelMember(
            channel_name=channel_name, user_id=user.id, role=new_member.role
        )

        self.db.add(member)
        self.db.commit()

    def get_package_members(self, channel_name, package_name):
        return (
            self.db.query(PackageMember)
            .join(User)
            .filter(User.username.isnot(None))
            .filter(PackageMember.channel_name == channel_name)
            .filter(PackageMember.package_name == package_name)
            .all()
        )

    def get_package_member(self, channel_name, package_name, username):
        return (
            self.db.query(PackageMember)
            .join(User)
            .filter(PackageMember.channel_name == channel_name)
            .filter(PackageMember.package_name == package_name)
            .filter(User.username == username)
            .one_or_none()
        )

    def create_package_member(self, channel_name, package_name, new_member):
        user = self.get_user_by_username(new_member.username)
        if user is None:
            raise KeyError(f"User '{new_member.username}' not found.")

        member = PackageMember(
            channel_name=channel_name,
            package_name=package_name,
            user_id=user.id,
            role=new_member.role,
        )

        self.db.add(member)
        self.db.commit()

    def get_api_keys_with_members(self, user_id, api_key_id=None):
        user_role_api_keys = (
            self.db.query(ApiKey)
            .filter(ApiKey.owner_id == user_id)
            .filter(ApiKey.user_id == user_id)
            .filter(~ApiKey.deleted)
        )

        custom_role_api_keys = (
            self.db.query(ApiKey, PackageMember, ChannelMember)
            .filter(ApiKey.owner_id == user_id)
            .filter(~ApiKey.deleted)
            .outerjoin(Profile, ApiKey.user_id == Profile.user_id)
            .filter(Profile.name.is_(None))
            .outerjoin(ChannelMember, ChannelMember.user_id == ApiKey.user_id)
            .outerjoin(PackageMember, PackageMember.user_id == ApiKey.user_id)
        )

        if api_key_id:
            user_role_api_keys = user_role_api_keys.filter(ApiKey.key == api_key_id)
            custom_role_api_keys = custom_role_api_keys.filter(ApiKey.key == api_key_id)

        user_role_api_keys = user_role_api_keys.all()
        custom_role_api_keys = custom_role_api_keys.all()

        return user_role_api_keys, custom_role_api_keys

    def get_package_api_keys(self, user_id):
        return (
            self.db.query(PackageMember, ApiKey)
            .join(User, PackageMember.user_id == User.id)
            .join(ApiKey, ApiKey.user_id == User.id)
            .filter(ApiKey.owner_id == user_id)
            .filter(~ApiKey.deleted)
            .all()
        )

    def get_channel_api_keys(self, user_id):
        return (
            self.db.query(ChannelMember, ApiKey)
            .join(User, ChannelMember.user_id == User.id)
            .join(ApiKey, ApiKey.user_id == User.id)
            .filter(ApiKey.owner_id == user_id)
            .filter(~ApiKey.deleted)
            .all()
        )

    def create_api_key(self, user_id, api_key: rest_models.BaseApiKey, key):
        owner = self.get_user(user_id)
        # if no roles are passed, create an API key with the same permissions as user
        if not api_key.roles:
            user = owner
        else:
            user = User(id=uuid.uuid4().bytes)
            self.db.add(user)

        db_api_key = ApiKey(
            key=key,
            description=api_key.description,
            time_created=date.today(),
            expire_at=api_key.expire_at,
            user=user,
            owner=owner,
        )

        self.db.add(db_api_key)
        if api_key.roles is not None:
            for role in api_key.roles:
                if role.package:
                    package_member = (
                        self.db.query(PackageMember)
                        .filter_by(
                            user=user,
                            channel_name=role.channel,
                            package_name=role.package,
                        )
                        .one_or_none()
                    )
                    if not package_member:
                        package_member = PackageMember(
                            user=user,
                            channel_name=role.channel,
                            package_name=role.package,
                            role=role.role,
                        )
                        self.db.add(package_member)
                    else:
                        package_member.role = role.role
                else:
                    channel_member = (
                        self.db.query(ChannelMember)
                        .filter_by(user=user, channel_name=role.channel)
                        .one_or_none()
                    )
                    if not channel_member:
                        channel_member = ChannelMember(
                            user=user, channel_name=role.channel, role=role.role
                        )
                        self.db.add(channel_member)
                    else:
                        channel_member.role = role.role

        self.db.commit()

        return db_api_key

    def get_api_key(self, key):
        return self.db.get(ApiKey, key)

    def create_version(
        self,
        channel_name,
        package_name,
        package_format,
        platform,
        version,
        build_number,
        build_string,
        filename,
        info,
        uploader_id,
        size,
        upsert: bool = False,
    ):
        # hold a lock on the package
        package = (  # noqa
            self.db.query(Package)
            .with_for_update()
            .filter(Package.channel_name == channel_name)
            .filter(Package.name == package_name)
            .join(PackageVersion)
            .filter(PackageVersion.package_format == package_format)
            .filter(PackageVersion.platform == platform)
        ).first()

        existing_versions = (
            self.db.query(PackageVersion)
            .filter(PackageVersion.channel_name == channel_name)
            .filter(PackageVersion.package_name == package_name)
            .filter(PackageVersion.package_format == package_format)
            .filter(PackageVersion.platform == platform)
            .filter(PackageVersion.version == version)
            .filter(PackageVersion.build_number == build_number)
            .filter(PackageVersion.build_string == build_string)
        )
        package_version = existing_versions.one_or_none()

        if not package_version:
            all_existing_versions = (
                self.db.query(PackageVersion)
                .filter(PackageVersion.channel_name == channel_name)
                .filter(PackageVersion.package_name == package_name)
                .order_by(PackageVersion.version_order.asc())
            ).all()

            version_order = 0

            if all_existing_versions:
                new_version = versionorder.VersionOrder(version)
                is_newer = False
                for v in all_existing_versions:
                    # type checker justly complains that v.version could be None
                    # ignoring it before attempting true fix
                    other = versionorder.VersionOrder(v.version)  # type: ignore
                    is_newer = other < new_version or (
                        other == new_version and v.build_number < build_number
                    )
                    if is_newer:
                        break
                version_order = (
                    v.version_order if is_newer else v.version_order + 1  # type: ignore
                )

                (
                    self.db.query(PackageVersion)
                    .filter(PackageVersion.channel_name == channel_name)
                    .filter(PackageVersion.package_name == package_name)
                    .filter(PackageVersion.package_format == package_format)
                    .filter(PackageVersion.platform == platform)
                    .filter(PackageVersion.version_order >= version_order)
                    .update(
                        {"version_order": PackageVersion.version_order + 1},
                    )
                )

            package_version = PackageVersion(
                id=uuid.uuid4().bytes,
                channel_name=channel_name,
                package_name=package_name,
                package_format=package_format,
                platform=platform,
                version=version,
                build_number=build_number,
                build_string=build_string,
                filename=filename,
                info=info,
                version_order=version_order,
                uploader_id=uploader_id,
                size=size,
            )

            self.db.add(package_version)
            logger.debug(
                f"adding package {package_name} version {version} to "
                + f"channel {channel_name}",
            )

        elif upsert:
            existing_versions.update(
                {
                    "filename": filename,
                    "info": info,
                    "uploader_id": uploader_id,
                    "time_modified": datetime.utcnow(),
                    "size": size,
                },
                synchronize_session="evaluate",
            )
        else:
            raise IntegrityError("duplicate package version", "", "")

        self.db.commit()

        return package_version

    def get_package_versions(
        self,
        package,
        time_created_ge: datetime = None,
        version_match_str: str = None,
        skip: int = 0,
        limit: int = -1,
    ):
        ApiKeyProfile = aliased(Profile)

        query = (
            self.db.query(PackageVersion, Profile, ApiKeyProfile)
            .outerjoin(Profile, Profile.user_id == PackageVersion.uploader_id)
            .outerjoin(ApiKey, ApiKey.user_id == PackageVersion.uploader_id)
            .outerjoin(ApiKeyProfile, ApiKey.owner_id == ApiKeyProfile.user_id)
            .filter(PackageVersion.channel_name == package.channel_name)
            .filter(PackageVersion.package_name == package.name)
            .order_by(PackageVersion.version_order.asc())
        )
        if time_created_ge:
            query = query.filter(PackageVersion.time_created >= time_created_ge)

        if version_match_str:
            if version_match:
                query = query.filter(
                    version_match(PackageVersion.version, version_match_str)
                )
            else:
                raise NotImplementedError(
                    "Quetz Database extension not loaded. Compile and configure "
                    "database_plugin_path correctly for support!"
                )

        if limit < 0:
            return query.all()
        else:
            return get_paginated_result(query, skip, limit)

    def get_package_version_by_filename(
        self, channel_name: str, package_name: str, filename: str, platform: str
    ) -> Optional[PackageVersion]:
        query = (
            self.db.query(PackageVersion)
            .filter(PackageVersion.channel_name == channel_name)
            .filter(PackageVersion.package_name == package_name)
            .filter(PackageVersion.filename == filename)
            .filter(PackageVersion.platform == platform)
        )

        return query.one_or_none()

    def is_active_platform(self, channel_name: str, platform: str):
        if platform == 'noarch':
            return True

        return (
            self.db.query(PackageVersion)
            .filter(PackageVersion.channel_name == channel_name)
            .filter(PackageVersion.platform == platform)
            .count()
            > 0
        )

    def get_package_infos(self, channel_name: str, subdir: str):
        # Returns iterator
        return (
            self.db.query(
                PackageVersion.filename,
                PackageVersion.info,
                PackageVersion.package_format,
                PackageVersion.time_modified,
            )
            .filter(PackageVersion.channel_name == channel_name)
            .filter(PackageVersion.platform == subdir)
            .order_by(PackageVersion.filename)
        )

    def get_channel_datas(self, channel_name: str):
        # Returns iterator
        return (
            self.db.query(Package.name, Package.channeldata)
            .filter(Package.channel_name == channel_name)
            .order_by(Package.name)
        )

    def assert_size_limits(self, channel_name: str, size: int):
        channel_size, channel_size_limit = (
            self.db.query(Channel.size, Channel.size_limit)
            .filter(Channel.name == channel_name)
            .one()
        )

        if channel_size_limit is not None:
            allowed = (channel_size + size) <= channel_size_limit

            if not allowed:
                raise errors.QuotaError(
                    f"{channel_name} is above quota of {channel_size_limit} bytes"
                )

    def update_channel_size(self, channel_name: str):
        channel_size = (
            self.db.query(func.sum(PackageVersion.size).label('size'))
            .filter(PackageVersion.channel_name == channel_name)
            .scalar()
        )

        if channel_size is None:
            channel_size = 0

        self.update_channel(channel_name, {"size": channel_size})

    def create_user_with_role(self, user_name: str, role: Optional[str] = None):
        """
        create a user without a profile or return a user if already exists
        and replace role
        """
        user = self.db.query(User).filter(User.username == user_name).one_or_none()
        if not user:
            user = User(id=uuid.uuid4().bytes, username=user_name, role=role)
            self.db.add(user)

        if role:
            user.role = role
        self.db.commit()
        return user

    def create_user_with_profile(
        self,
        username: str,
        provider: str,
        identity_id: str,
        name: str,
        avatar_url: str,
        role: Optional[str],
        exist_ok: bool = False,
        emails: Optional[List['auth_base.Email']] = None,
    ):
        """create a user with profile and role

        :param exist_ok: flag to check whether the user should be reused if exists
          or raise an error
        """

        # TODO: check that username comes from the right id provider
        user = self.db.query(User).filter(User.username == username).one_or_none()

        if not exist_ok:
            if user:
                raise IntegrityError(f"User {username} exists", "", "")

            # check if any email already registered
            if emails:
                for e in emails:
                    user_email = (
                        self.db.query(Email)
                        .filter(Email.email == e["email"])
                        .one_or_none()
                    )
                    if user_email:
                        raise IntegrityError(
                            f"User {username} already registered "
                            f"with email {user_email.email}",
                            "",
                            "",
                        )

        if not user:
            user = User(id=uuid.uuid4().bytes, username=username, role=role)

        identity = Identity(
            provider=provider,
            identity_id=identity_id,
            username=username,
        )

        profile = Profile(name=name, avatar_url=avatar_url)

        user.identities.append(identity)

        if emails:
            user_emails = []
            for email in emails:
                # we only store verified emails
                if not email["verified"]:
                    continue
                user_email = Email(
                    provider=provider,
                    identity_id=identity_id,
                    email=email["email"],
                    verified=email["verified"],
                    primary=email["primary"],
                )
                user_emails.append(user_email)

            user.emails.extend(user_emails)

        user.profile = profile
        self.db.add(user)
        self.db.commit()
        self.db.refresh(user)

        return user

    def get_jobs(
        self,
        states: Optional[List[JobStatus]] = None,
        skip: int = 0,
        limit: int = -1,
        owner_id: Optional[bytes] = None,
    ):
        jobs = self.db.query(Job)

        if states:
            jobs = jobs.filter(Job.status.in_(states))
        if owner_id:
            jobs = jobs.filter(Job.owner_id == owner_id)
        jobs = jobs.order_by(Job.id)

        return get_paginated_result(jobs, skip, limit)

    def get_job(self, job_id: int):
        job = self.db.query(Job).filter(Job.id == job_id).one_or_none()
        return job

    def get_tasks(
        self,
        job_id: int,
        states: Optional[List[TaskStatus]] = None,
        skip: int = 0,
        limit: int = -1,
    ):
        tasks = self.db.query(Task).filter(Task.job_id == job_id)

        if states:
            tasks = tasks.filter(Task.status.in_(states))

        return get_paginated_result(tasks, skip, limit)

    def create_job(self, user_id, job_model):
        serialized = job_model.manifest.encode('ascii')
        job = Job(
            owner_id=user_id,
            manifest=serialized,
            items_spec=job_model.items_spec,
            start_at=job_model.start_at,
            repeat_every_seconds=job_model.repeat_every_seconds,
        )
        self.db.add(job)
        self.db.commit()
        return job

    def incr_download_count(
        self,
        channel: str,
        filename: str,
        platform: str,
        timestamp: Optional[datetime] = None,
        incr: int = 1,
    ):
        metric_name = "download"

        self.db.query(PackageVersion).filter(
            PackageVersion.channel_name == channel
        ).filter(PackageVersion.filename == filename).filter(
            PackageVersion.platform == platform
        ).update(
            {PackageVersion.download_count: PackageVersion.download_count + incr}
        )

        if timestamp is None:
            timestamp = datetime.utcnow()

        all_values = []
        for interval in IntervalType:
            values = {
                'channel_name': channel,
                'platform': platform,
                'metric_name': metric_name,
                'filename': filename,
                "timestamp": date_trunc(interval, timestamp),
                "period": interval,
                "count": incr,
            }

            all_values.append(values)

        index_elements = [
            'channel_name',
            'platform',
            'filename',
            'metric_name',
            'period',
            'timestamp',
        ]

        stmt = Upsert(
            PackageVersionMetric.__table__,
            all_values,
            index_elements,
            PackageVersionMetric.count,
            incr=incr,
        )

        self.db.execute(stmt)

        self.db.commit()

    def get_package_version_metrics(
        self,
        package_version_id,
        period,
        metric_name,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        fill_zeros: bool = False,
    ):
        m = PackageVersionMetric
        v = PackageVersion

        q = (
            self.db.query(m)
            .join(
                v,
                and_(
                    v.platform == m.platform,
                    v.channel_name == m.channel_name,
                    m.filename == v.filename,
                ),
            )
            .filter(v.id == package_version_id)
            .filter(m.period == period)
            .filter(m.metric_name == metric_name)
            .order_by(m.timestamp)
        )

        if start:
            q = q.filter(m.timestamp >= start)

        if end:
            q = q.filter(m.timestamp < end)

        items = q.all()

        if fill_zeros:

            def factory():
                return PackageVersionMetric(
                    count=0,
                    metric_name=metric_name,
                    period=period,
                )

            timestamps: Dict[datetime, PackageVersionMetric] = defaultdict(factory)

            for d in items:
                timestamps[d.timestamp] = d

            start = start or items[0].timestamp
            end = end or items[-1].timestamp

            first = round_timestamp(start, period)
            last = round_timestamp(end, period)

            metrics = []
            while first <= last:
                item = timestamps[first]
                item.timestamp = first
                metrics.append(item)
                first = next_timestamp(first, period)
            return metrics
        else:
            return items

    def get_channel_metrics(
        self,
        channel_name,
        period,
        metric_name,
        platform: Optional[str] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ):
        m = PackageVersionMetric

        q = (
            self.db.query(m)
            .filter(m.channel_name == channel_name)
            .filter(m.period == period)
            .filter(m.metric_name == metric_name)
        )

        if platform:
            q = q.filter(m.platform == platform)

        q = q.order_by(m.platform, m.filename, m.timestamp)

        if start:
            q = q.filter(m.timestamp >= start)

        if end:
            q = q.filter(m.timestamp < end)

        rows_per_filename = groupby(q, key=lambda row: f"{row.platform}/{row.filename}")

        return {
            filename: {"series": list(group)} for filename, group in rows_per_filename
        }



from datetime import datetime, timedelta
from unittest.mock import ANY, Mock

import pytest

from quetz.dao import Dao
from quetz.metrics.db_models import IntervalType, next_timestamp, round_timestamp
from quetz.metrics.tasks import synchronize_metrics_from_mirrors


def test_round_timestamp():
    timestamp = datetime.fromisoformat("2020-02-03T04:05")
    rounded = datetime.fromisoformat("2020-02-03T00:00")
    period = IntervalType.day
    assert round_timestamp(timestamp, period) == rounded

    period = IntervalType.hour
    rounded = datetime(2020, 2, 3, 4, 00)
    assert round_timestamp(timestamp, period) == rounded


def test_next_timestamp():
    t = datetime.fromisoformat("2020-02-03T04:05")

    T = IntervalType

    assert next_timestamp(t, T.hour) == datetime(2020, 2, 3, 5, 5)
    assert next_timestamp(t, T.day) == datetime(2020, 2, 4, 4, 5)
    assert next_timestamp(t, T.month) == datetime(2020, 3, 3, 4, 5)
    assert next_timestamp(t, T.year) == datetime(2021, 2, 3, 4, 5)

    # special cases
    assert next_timestamp(datetime(2020, 12, 31, 23, 55), T.hour) == datetime(
        2021, 1, 1, 0, 55
    )
    assert next_timestamp(datetime(2020, 12, 31, 23, 55), T.month) == datetime(
        2021, 1, 31, 23, 55
    )
    with pytest.raises(ValueError):
        assert next_timestamp(datetime(2020, 1, 31, 23, 55), T.month) == datetime(
            2021, 2, 28, 23, 55
        )


def test_get_download_count(auth_client, public_channel, package_version, db, dao: Dao):
    timestamps = [
        "2020-01-05T21:01",
        "2020-01-06T22:10",
        "2020-02-18T10:10",
    ]

    month_day = []
    for t in timestamps:
        dt = datetime.fromisoformat(t)
        dao.incr_download_count(
            public_channel.name,
            package_version.filename,
            package_version.platform,
            timestamp=dt,
        )
        month_day.append((dt.month, dt.day))

    endpoint_url = (
        f"/metrics/channels/{public_channel.name}/"
        f"packages/{package_version.package_name}/"
        f"versions/{package_version.platform}/{package_version.filename}"
    )

    response = auth_client.get(endpoint_url)
    assert response.status_code == 200

    assert response.json() == {
        "server_timestamp": ANY,
        "period": "D",
        "metric_name": "download",
        "total": 3,
        "series": [
            {"timestamp": f"2020-{m:02}-{d:02}T00:00:00", "count": 1}
            for m, d in month_day
        ],
    }

    response = auth_client.get(endpoint_url + "?start=2020-01-05T10:00")
    assert response.status_code == 200

    assert response.json() == {
        "server_timestamp": ANY,
        "period": "D",
        "metric_name": "download",
        "total": 2,
        "series": [
            {"timestamp": f"2020-{m:02}-{d:02}T00:00:00", "count": 1}
            for m, d in month_day[1:]
        ],
    }

    # fill zeros
    response = auth_client.get(
        endpoint_url + "?start=2020-01-05T20:00"
        "&end=2020-01-05T22:00"
        "&period=H"
        "&fill_zeros=true"
    )
    assert response.status_code == 200

    assert response.json() == {
        "server_timestamp": ANY,
        "period": "H",
        "metric_name": "download",
        "total": 1,
        "series": [
            {"timestamp": "2020-01-05T20:00:00", "count": 0},
            {"timestamp": "2020-01-05T21:00:00", "count": 1},
            {"timestamp": "2020-01-05T22:00:00", "count": 0},
        ],
    }

    response = auth_client.get(endpoint_url + "?period=M")
    assert response.status_code == 200

    assert response.json() == {
        "server_timestamp": ANY,
        "period": "M",
        "metric_name": "download",
        "total": 3,
        "series": [
            {"timestamp": f"2020-{m:02}-01T00:00:00", "count": c}
            for m, c in [(1, 2), (2, 1)]
        ],
    }


@pytest.fixture
def package_version_factory(channel_name, user, dao, public_package, db):
    package_format = "tarbz2"
    package_info = "{}"
    package_name = public_package.name

    versions = []

    def factory(version, build_str=0, platform="linux-64"):
        filename = f"{package_name}-{version}-{build_str}.tar.bz2"

        version = dao.create_version(
            channel_name,
            package_name,
            package_format,
            "linux-64",
            version,
            build_str,
            str(build_str),
            str(filename),
            package_info,
            user.id,
            size=11,
        )
        versions.append(version)
        return version

    yield factory

    for version in versions:
        db.delete(version)
    db.commit()


def test_get_channel_download_count(
    auth_client, public_channel, package_version_factory, db, dao: Dao
):
    versions = [package_version_factory(str(i)) for i in range(3)]

    now = datetime.utcnow()

    for v in versions:
        dao.incr_download_count(
            public_channel.name,
            v.filename,
            v.platform,
            timestamp=now,
        )

    endpoint_url = f"/metrics/channels/{public_channel.name}"

    response = auth_client.get(endpoint_url)
    assert response.status_code == 200

    expected = {
        "server_timestamp": ANY,
        "metric_name": "download",
        "period": "D",
        "packages": {
            v.platform
            + '/'
            + v.filename: {
                "series": [
                    {
                        "timestamp": now.replace(
                            minute=0, second=0, microsecond=0, hour=0
                        ).isoformat(),
                        "count": 1,
                    }
                ]
            }
            for v in versions
        },
    }

    assert response.json() == expected


@pytest.fixture
def channel_mirror(public_channel, dao: Dao):
    mirror_url = "http://mirror_server/get/my-mirror"
    api_endpoint = "http://mirror_server/api/my-mirror"
    metrics_endpoint = "http://mirror_server/metrics/channels/my-mirror"
    return dao.create_channel_mirror(
        public_channel.name, mirror_url, api_endpoint, metrics_endpoint
    )


def test_synchronize_metrics_without_mirrors(public_channel, package_version, dao: Dao):
    session = Mock()

    synchronize_metrics_from_mirrors(public_channel.name, dao, session)

    metrics = dao.get_package_version_metrics(
        package_version.id, IntervalType.hour, "download"
    )

    assert not metrics
    session.get.assert_not_called()


def test_synchronize_metrics_with_mirrors(
    public_channel, package_version, channel_mirror, dao: Dao
):
    timestamp = datetime(2020, 10, 1, 5, 0)
    first_sync_time = timestamp - timedelta(days=1)
    sync_time = timestamp + timedelta(hours=1)

    session = Mock()
    session.get.return_value.json.return_value = {
        "server_timestamp": first_sync_time.isoformat(),
        "packages": {},
    }
    session.get.return_value.status_code = 200
    synchronize_metrics_from_mirrors(
        public_channel.name, dao, session, now=first_sync_time
    )

    metrics = dao.get_package_version_metrics(
        package_version.id, IntervalType.hour, "download"
    )

    assert not metrics
    session.get.assert_called_with(
        "http://mirror_server/metrics/channels/my-mirror"
        f"?period=H&end={first_sync_time.isoformat()}"
    )
    session.reset_mock()

    session.get.return_value.json.return_value = {
        "server_timestamp": sync_time.isoformat(),
        "packages": {
            f"{package_version.platform}/{package_version.filename}": {
                "series": [
                    {"timestamp": timestamp.isoformat(), "count": 2},
                ]
            }
        },
    }

    synchronize_metrics_from_mirrors(public_channel.name, dao, session, sync_time)

    for p in IntervalType:
        metrics = dao.get_package_version_metrics(package_version.id, p, "download")
        assert len(metrics) == 1
        assert metrics[0].count == 2

    session.get.assert_called_with(
        "http://mirror_server/metrics/channels/my-mirror"
        f"?period=H&start={first_sync_time.isoformat()}&end={sync_time.isoformat()}"
    )

    session.reset_mock()
    hour = timedelta(hours=1)
    session.get.return_value.json.return_value = {
        "server_timestamp": sync_time.replace(minute=5).isoformat(),
        "packages": {
            f"{package_version.platform}/{package_version.filename}": {
                "series": [
                    {
                        "timestamp": (timestamp + hour).isoformat(),
                        "count": 1,
                    },
                ]
            }
        },
    }

    third_sync_time = sync_time + hour

    synchronize_metrics_from_mirrors(public_channel.name, dao, session, third_sync_time)
    session.get.assert_called_with(
        "http://mirror_server/metrics/channels/my-mirror"
        f"?period=H&start={sync_time.isoformat()}&end={third_sync_time.isoformat()}"
    )

    metrics = dao.get_package_version_metrics(
        package_version.id, IntervalType.day, "download"
    )
    assert len(metrics) == 1
    assert metrics[0].count == 3



import logging
import pickle
import uuid
from datetime import datetime
from typing import Optional

from importlib_metadata import entry_points as get_entry_points
from pydantic import BaseModel, ConfigDict, Field, field_validator

from . import handlers
from .models import JobStatus, TaskStatus

logger = logging.getLogger("quetz")


def parse_job_manifest(function_name):
    """validate and parse job function name from a string

    Examples:

    parse_job_manifest("some_function")

       returns one of the built-in functions registered in quetz.jobs.handlers modules

    parse_job_manifest("plugin:function_name")

       returns a function from a moduled registered with plugin's quetz.jobs entrypoint

    parse_job_manifest("non_existent_function")

       raises ValueError for unknown functions

    """
    paths = function_name.split(":")

    if len(paths) == 2:
        plugin_name, job_name = paths
        entry_points = tuple(
            get_entry_points().select(group='quetz.jobs', name=plugin_name)
        )
        if not entry_points:
            raise ValueError(
                f"invalid function {function_name}: "
                f"plugin {plugin_name} not installed"
            )
        job_module = entry_points[0].load()
        try:
            return getattr(job_module, job_name)
        except AttributeError:
            raise ValueError(
                f"invalid function '{job_name}' name in plugin '{plugin_name}'"
            )
    elif len(paths) == 1:
        try:
            return handlers.JOB_HANDLERS[function_name]
        except KeyError:
            raise ValueError(
                f"invalid function {function_name}: no such built-in function,"
                " please provide plugin name"
            )
    else:
        raise ValueError(f"invalid function {function_name} - could not parse")


def parse_job_name(v):
    try:
        return v.decode("ascii")
    except UnicodeDecodeError:
        pass

    # try unpickling

    try:
        func = pickle.loads(v)
        return f"{func.__module__}:{func.__name__}"
    except pickle.UnpicklingError:
        raise ValueError("could not parse manifest")
    except ModuleNotFoundError as e:
        logger.error(f"job function not found: could not import module {e.name}")
        return f"{e.name}:undefined"


class JobBase(BaseModel):
    """New job spec"""

    manifest: str = Field(None, title='Name of the function')

    start_at: Optional[datetime] = Field(
        None, title="date and time the job should start, if None it starts immediately"
    )
    repeat_every_seconds: Optional[int] = Field(
        None,
        title=(
            "interval in seconds at which the job should be repeated, "
            "if None it is a one-off job"
        ),
    )

    @field_validator("manifest", mode="before")
    @classmethod
    def validate_job_name(cls, function_name):
        if isinstance(function_name, bytes):
            return parse_job_name(function_name)

        parse_job_manifest(function_name)

        return function_name.encode('ascii')


class JobCreate(JobBase):
    """Create job spec"""

    items_spec: str = Field(..., title='Item selector spec')


class JobUpdateModel(BaseModel):
    """Modify job spec items (status and items_spec)"""

    items_spec: str = Field(None, title='Item selector spec')
    status: JobStatus = Field(None, title='Change status')
    force: bool = Field(False, title="force re-running job on all matching packages")


class Job(JobBase):
    id: int = Field(None, title='Unique id for job')
    owner_id: uuid.UUID = Field(None, title='User id of the owner')

    created: datetime = Field(None, title='Created at')

    status: JobStatus = Field(None, title='Status of the job (running, paused, ...)')

    items_spec: Optional[str] = Field(None, title='Item selector spec')
    model_config = ConfigDict(from_attributes=True)


class Task(BaseModel):
    id: int = Field(None, title='Unique id for task')
    job_id: int = Field(None, title='ID of the parent job')
    package_version: dict = Field(None, title='Package version')
    created: datetime = Field(None, title='Created at')
    status: TaskStatus = Field(None, title='Status of the task (running, paused, ...)')

    @field_validator("package_version", mode="before")
    @classmethod
    def convert_package_version(cls, v):
        if v:
            return {'filename': v.filename, 'id': uuid.UUID(bytes=v.id).hex}
        else:
            return {}

    model_config = ConfigDict(from_attributes=True)



import sys
import uuid
from typing import Dict, List, Optional, Type, Union

from fastapi import APIRouter, Depends, Request, Response
from sqlalchemy.sql.sqltypes import Boolean
from starlette.responses import RedirectResponse

from quetz.authorization import ServerRole
from quetz.config import Config
from quetz.dao import Dao
from quetz.deps import get_config, get_dao

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict

from . import auth_dao


class Email(TypedDict):
    email: str
    verified: Boolean
    primary: Boolean


class UserProfile(TypedDict):
    id: str
    name: str
    avatar_url: str
    login: str
    emails: List[Email]


class UserName(TypedDict):
    username: str


class UserDict(UserName, total=False):
    # profile and auth_state keys are optional
    profile: UserProfile
    auth_state: Dict[str, str]


AuthReturnType = Optional[Union[str, UserDict]]


class BaseAuthenticationHandlers:
    """Handlers for authenticator endpoints.

    Subclasses MUST implement:

    - login method
    - authorize_methods: provide allowed HTTP methods for authorize endpoint

    SHOULD override:

    - _authenticate method - to extract data and past it to
      BaseAuthenticator.authenticate method
    """

    # list of methods that /authorize endpoint can be requested with
    authorize_methods: List[str]

    def __init__(self, authenticator, app=None):
        self.authenticator = authenticator

        # dependency_overrides_provider kwarg is needed for unit test
        self.router = APIRouter(
            prefix=f"/auth/{authenticator.provider}", dependency_overrides_provider=app
        )
        self.router.add_api_route("/login", self.login, methods=["GET"])
        self.router.add_api_route("/enabled", self.enabled, methods=["GET"])

        self.router.add_api_route(
            "/authorize",
            self.authorize,
            methods=self.authorize_methods,
            name=f"authorize_{authenticator.provider}",
        )

    async def login(self, request: Request):
        """{prefix}/login endpoint
        First entry point to trigger login process"""
        raise NotImplementedError("method login must be implemented in subclasses")

    async def enabled(self):
        """{prefix}/enabled endpoint
        Used by frontend to show the login button."""

        return self.authenticator.is_enabled

    async def authorize(
        self,
        request: Request,
        dao: Dao = Depends(get_dao),
        config: Config = Depends(get_config),
    ):
        """{prefix}/authorize endpoint
        Entry point for user submitted data or callback for oauth applications.

        To configure HTTP method that this endpoint will handle, set authorize_methods.
        """

        user_dict = await self._authenticate(request, dao, config)

        if user_dict is None:
            return Response("login failed")

        if isinstance(user_dict, str):
            # wrap string in a dictionary
            user_data: UserDict = {"username": user_dict}
        else:
            user_data = user_dict

        default_profile: UserProfile = {
            "login": user_data["username"],
            "id": user_data["username"],
            "name": user_data["username"],
            "avatar_url": "",
            "emails": [],
        }

        profile: UserProfile = user_data.get("profile", default_profile)

        role = await self.authenticator.user_role(request, profile)
        default_channel = await self.authenticator.user_channels(request, profile)

        user = auth_dao.get_user_by_identity(
            dao, self.authenticator.provider, profile, config, role, default_channel
        )

        user_id = str(uuid.UUID(bytes=user.id))

        request.session['user_id'] = user_id
        request.session['identity_provider'] = self.authenticator.provider
        request.session.update(user_data.get("auth_state", {}))

        # use 303 code so that the method is always changed to GET
        resp = RedirectResponse('/', status_code=303)

        return resp

    async def _authenticate(self, request, dao, config) -> AuthReturnType:
        """wrapper around `authenticate` method of the Authenticator subclasses

        mainly used to extract data from request."""

        user_dict = await self.authenticator.authenticate(
            request, data=None, dao=dao, config=config
        )
        return user_dict


class BaseAuthenticator:
    """Base class for authenticators.

    Subclasses MUST implement:

    - :py:meth:`authenticate`

    Subclasses SHOULD implement:

    - :py:meth:`configure`
    - :py:meth:`validate_token`

    """

    provider = "base"
    handler_cls: Type[BaseAuthenticationHandlers] = BaseAuthenticationHandlers

    is_enabled = False

    default_role: Optional[str] = None
    default_channel: Optional[str] = None

    @property
    def router(self):
        return self.handler.router

    def __init__(self, config: Config, provider=None, app=None):
        if provider is not None:
            self.provider = str(provider)
        self.handler = self.handler_cls(self, app)

        self.configure(config)

    def configure(self, config):
        """configure authenticator and set is_enabled=True"""

        # use the config to configure default role and
        # whether to create a default channel
        if config.configured_section("users"):
            self.default_role = config.users_default_role
            self._admins = config.users_admins
            self._maintainers = config.users_maintainers
            self._members = config.users_members
            self.create_default_channel = config.users_create_default_channel
        else:
            self.default_role = None
            self.create_default_channel = False
            self._admins = []
            self._maintainers = []
            self._members = []

    async def validate_token(self, token):
        "check token validity"
        return True

    async def user_role(self, request: Request, profile: UserProfile) -> Optional[str]:
        """return default role of the new user"""
        login = profile['login']
        user_str = f"{self.provider}:{login}"

        if user_str in self._admins:
            return ServerRole.OWNER
        elif user_str in self._maintainers:
            return ServerRole.MAINTAINER
        elif user_str in self._members:
            return ServerRole.MEMBER
        else:
            return self.default_role

    async def user_channels(
        self, request: Request, profile: UserProfile
    ) -> Optional[List[str]]:
        """user channel"""
        if self.create_default_channel:
            return [profile["login"]]
        else:
            return None

    async def authenticate(
        self, request, data=None, dao=None, config=None, **kwargs
    ) -> AuthReturnType:
        """Authentication user with the data submitted.

        This method should return:

        - ``None`` if the authentication failed,
        - string with username for successful authentication
        - or a dictionary with keys ``username``, ``profile`` (user profile data),
          ``auth_state`` (extra authentication state to be stored in the browser
          session).

        """

        raise NotImplementedError("subclasses need to implement authenticate")


class FormHandlers(BaseAuthenticationHandlers):
    authorize_methods = ["POST"]

    async def login(self, request: Request):
        redirect_uri = request.url_for(f'authorize_{self.authenticator.provider}')
        data = f"""
<html><body><h1>Login Page</h1>
<form method="post" action="{redirect_uri}">
  <label>username:
    <input name="username" autocomplete="name">
  </label>
  <label>password:
    <input name="password" type="password">
  </label>
  <button>Submit</button>
</form>
</body></html>"""
        return Response(content=data, media_type="text/html")

    async def _authenticate(self, request, dao, config):
        """Wrapper around `authenticate` method of the Authenticator subclasses.

        Extracts form data from request."""

        data = await request.form()

        user_dict = await self.authenticator.authenticate(
            request, data=data, dao=dao, config=config
        )

        return user_dict


class SimpleAuthenticator(BaseAuthenticator):
    """A demo of a possible implementation. It redirects users to a
    simple HTML form where they can type their username and password.

    Note: Consider this an example. In your production setting you would
    probably want to redirect to your custom login page. Make sure that
    this page submits data to ``/auth/{provider}/authorize`` endpoint."""

    provider = "simple"
    handler_cls = FormHandlers

    def configure(self, config):
        self.is_enabled = True

        super().configure(config)

    async def authenticate(
        self, request: Request, data=None, dao=None, config=None, **kwargs
    ) -> Optional[str]:
        if data is None:
            return None
        elif data["username"] == data["password"]:
            return data['username']
        else:
            return None



import getpass
from unittest import mock

import pytest
from fastapi import Request

from quetz.authentication.pam import PAMAuthenticator


@pytest.fixture
def groups():
    return {}


@pytest.fixture
def config_extra(groups):
    return f"""
[pamauthenticator]
admin_groups = {groups.get('admins', [])}
maintainer_groups = {groups.get('maintainers', [])}
member_groups = {groups.get('members', [])}"""


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "groups,expected_role",
    [
        (
            {},
            None,
        ),
        (
            {"admins": ["usergroup"]},
            "owner",
        ),
        (
            {"admins": ["usergroup"], "maintainers": ["usergroup"]},
            "owner",
        ),
        (
            {"maintainers": ["usergroup"]},
            "maintainer",
        ),
        (
            {"members": ["usergroup"]},
            "member",
        ),
        pytest.param(
            {"members": ["missinggroup"]},
            None,
            id="missing-group",
        ),
    ],
)
async def test_user_role(config, expected_role):
    auth = PAMAuthenticator(config)
    request = Request(scope={"type": "http"})

    _group_ids = {"usergroup": 1001}
    _user_group_ids = {"quetzuser": [1001, 1002]}

    with mock.patch.multiple(
        auth,
        _get_group_id_by_name=lambda k: _group_ids[k],
        _get_user_group_ids=lambda k: _user_group_ids[k],
    ):
        role = await auth.user_role(request, {"login": "quetzuser"})

    if expected_role is None:
        assert role is None
    else:
        assert role == expected_role


@pytest.mark.asyncio
async def test_authenticate(config):
    auth = PAMAuthenticator(config)
    request = Request(scope={"type": "http"})
    current_user = getpass.getuser()

    result = await auth.authenticate(
        request, {"username": current_user, "password": "test"}
    )

    assert result is None  # authentication failed due to incorrect password
